"""Classes and functions for working with the Echo extension."""
#
# (C) Pywikibot team, 2014-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import pywikibot


@dataclass(eq=False)
class Notification:

    """A notification issued by the Echo extension.

    .. versionchanged:: 3.0.20190204
       The ``id`` attribute was renamed to ``event_id``, and its type
       changed from ``str`` to ``int``.

    .. deprecated:: 3.0.20190204
       The ``id`` attribute was retained temporarily for backward
       compatibility, but is deprecated and scheduled for removal.

    .. versionremoved:: 7.0
       The ``id`` attribute was removed.

    .. versionchanged:: 10.3
       The class is now defined using the ``@dataclass`` decorator to
       simplify internal initialization and improve maintainability.
    """

    site: pywikibot.site.BaseSite

    def __post_init__(self) -> None:
        """Initialize attributes for an empty Notification object.

        .. versionadded: 10.3
        """
        self.event_id: int | None = None
        self.type = None
        self.category = None
        self.timestamp: pywikibot.Timestamp | None = None
        self.page: pywikibot.Page | None = None
        self.agent: pywikibot.User | None = None
        self.read: pywikibot.Timestamp | bool | None = None
        self.content = None
        self.revid = None

    @classmethod
    def fromJSON(cls,  # noqa: N802
                 site: pywikibot.site.BaseSite,
                 data: dict[str, Any]) -> Notification:
        """Construct a Notification object from API JSON data.

        :param site: The pywikibot site object.
        :param data: The JSON data dictionary representing a
            notification.
        :return: An instance of Notification.
        """
        notif = cls(site)

        notif.event_id = int(data['id'])
        notif.type = data['type']
        notif.category = data['category']

        notif.timestamp = pywikibot.Timestamp.fromtimestampformat(
            data['timestamp']['mw'])

        try:
            notif.page = pywikibot.Page(site, data['title']['full'])
        except KeyError:
            notif.page = None

        try:
            notif.agent = pywikibot.User(site, data['agent']['name'])
        except KeyError:
            notif.agent = None

        try:
            notif.read = pywikibot.Timestamp.fromtimestampformat(data['read'])
        except KeyError:
            notif.read = False

        notif.content = data.get('*')
        notif.revid = data.get('revid')

        return notif

    def mark_as_read(self) -> bool:
        """Mark the notification as read.

        :return: True if the notification was successfully marked as
            read, else False.
        """
        if self.event_id is None:
            return False

        return self.site.notifications_mark_read(**{'list': self.event_id})



"""Family module for Wikifunction.

.. versionadded:: 8.4
"""
#
# (C) Pywikibot team, 2023
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from pywikibot import family


class Family(family.SingleSiteFamily, family.WikimediaFamily):

    """Family class for Wikifunctions."""

    name = 'wikifunctions'
    domain = 'www.wikifunctions.org'



"""Basic HTTP access interface.

This module handles communication between the bot and the HTTP threads.

This module is responsible for
    - Setting up a connection pool
    - Providing a (blocking) interface for HTTP requests
    - Translate site objects with query strings into URLs
    - URL-encoding all data
    - Basic HTTP error handling

This module creates and uses its own ``requests.Session`` object.
The session is closed if the module terminates. If required you can use
your own Session object passing it to the ``http.session`` variable::

    from pywikibot.comms import http
    session = requests.Session()
    http.session = session

To enable access via cookies, assign cookie handling class::

    session.cookies = http.cookie_jar

.. versionchanged:: 8.0
   Cookies are lazy loaded when logging to site.
"""
#
# (C) Pywikibot team, 2007-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import atexit
import codecs
import re
import sys
import threading
import traceback
from contextlib import suppress
from http import HTTPStatus, cookiejar
from string import Formatter
from urllib.parse import quote, urlparse
from warnings import warn

import requests

import pywikibot
from pywikibot import config, tools
from pywikibot.exceptions import (
    Client414Error,
    FatalServerError,
    Server504Error,
    ServerError,
)
from pywikibot.logging import critical, debug, error, log, warning
from pywikibot.tools import file_mode_checker


try:
    import requests_oauthlib
except ImportError as e:
    requests_oauthlib = e


class PywikibotCookieJar(cookiejar.LWPCookieJar):

    """CookieJar which create the filename and checks file permissions.

    .. versionadded:: 8.0
    .. versionchanged:: 10.2
       use `threading.Lock` in :meth:`load` and :meth`save` to be thread
       safe.
    """

    _lock = threading.Lock()  # Class-level lock shared across all instances

    def load(self, user: str = '', *args, **kwargs) -> None:
        """Loads cookies from a file.

        Insert the account name to the cookie filename, set the
        instance`s filename and load the cookies.

        :param user: account name to be part of the cookie filename.
        """
        _user = '-' + tools.as_filename(user) if user else ''
        self.filename = config.datafilepath(f'pywikibot{_user}.lwp')

        try:
            with self._lock:
                super().load(*args, **kwargs)
        except (cookiejar.LoadError, FileNotFoundError):
            debug(f'Loading cookies for user {user} failed.')
        else:
            debug(f'Loaded cookies for user {user} from file.')

    def save(self, *args, **kwargs) -> None:
        """Check the file mode and save cookies to a file.

        .. note:: *PywikibotCookieJar* must be loaded previously to set
           the filename.

        :raises ValueError: a filename was not supplied; :meth:`load`
            must be called first.
        """
        with self._lock:
            if self.filename:
                file_mode_checker(self.filename, create=True)
            super().save(*args, **kwargs)


#: global :class:`PywikibotCookieJar` instance.
cookie_jar = PywikibotCookieJar()
#: global :class:`requests.Session`.
session = requests.Session()
session.cookies = cookie_jar


def flush() -> None:  # pragma: no cover
    """Close the session object. This is called when the module terminates.

    .. versionchanged:: 8.1
       log the traceback and show the exception value in the critical
       message
    """
    log('Closing network session.')
    session.close()

    if hasattr(sys, 'last_type'):
        log(
            ''.join(
                traceback.format_exception(
                    sys.last_type,
                    value=sys.last_value,
                    tb=sys.last_traceback
                )
            )
        )
        critical(
            f'Exiting due to uncaught exception {sys.last_type.__name__}: '
            f'{sys.last_value}'
        )

    log('Network session closed.')


atexit.register(flush)

USER_AGENT_PRODUCTS = {
    'python': 'Python/' + '.'.join(str(i) for i in sys.version_info),
    'http_backend': 'requests/' + requests.__version__,
    'pwb': 'Pywikibot/' + pywikibot.__version__,
}


class _UserAgentFormatter(Formatter):

    """User-agent formatter to load version/revision only if necessary."""

    def get_value(self, key, args, kwargs):
        """Get field as usual except for version and revision."""
        # This is the Pywikibot version; also map it to {revision} at present.
        if key in ('version', 'revision'):
            return pywikibot.version.getversiondict()['rev']
        return super().get_value(key, args, kwargs)


_USER_AGENT_FORMATTER = _UserAgentFormatter()


def user_agent_username(username=None):
    """Reduce username to a representation permitted in HTTP headers.

    To achieve that, this function:

    - replaces spaces (' ') with '_'
    - encodes the username as 'utf-8' and if the username is not ASCII
    - URL encodes the username if it is not ASCII, or contains '%'
    """
    if not username:
        return ''

    username = username.replace(' ', '_')  # Avoid spaces or %20.
    try:
        username.encode('ascii')  # just test, but not actually use it
    except UnicodeEncodeError:
        username = quote(username.encode('utf-8'))
    else:
        # % is legal in the default $wgLegalTitleChars
        # This is so that ops know the real pywikibot will not
        # allow a useragent in the username to allow through a hand-coded
        # percent-encoded value.
        if '%' in username:
            username = quote(username)
    return username


def user_agent(site: pywikibot.site.BaseSite | None = None,
               format_string: str | None = '') -> str:
    """Generate the user agent string for a given site and format.

    :param site: The site for which this user agent is intended. May be
        None.
    :param format_string: The string to which the values will be added
        using str.format. Is using config.user_agent_format when it is
        empty.
    :return: The formatted user agent
    """
    values = USER_AGENT_PRODUCTS.copy()
    values.update(dict.fromkeys(['script', 'script_product'],
                                pywikibot.bot.calledModuleName()))
    values.update(dict.fromkeys(['family', 'code', 'lang', 'site'], ''))

    script_comments: list[str] = []
    if config.user_agent_description:
        script_comments.append(config.user_agent_description)

    username = ''
    if site:
        script_comments.append(str(site))

        # TODO: there are several ways of identifying a user, and username
        # is not the best for a HTTP header if the username isn't ASCII.
        if site.username():
            username = user_agent_username(site.username())
            script_comments.append('User:' + username)

        values.update({
            'family': site.family.name,
            'code': site.code,
            'lang': (site.lang if site.siteinfo.is_cached('lang')
                     else f'({site.code})'),
            'site': str(site),
        })

    values['username'] = username
    values['script_comments'] = '; '.join(script_comments)

    format_string = format_string or config.user_agent_format
    formatted = _USER_AGENT_FORMATTER.format(format_string, **values)

    # clean up after any blank components
    return formatted.replace('()', '').replace('  ', ' ').strip()


def fake_user_agent() -> str:
    """Return a fake user agent."""
    try:
        from fake_useragent import UserAgent
    except ImportError:
        raise ImportError(  # Actually complain when fake_useragent is missing.
            'fake_useragent must be installed to get fake UAs.')
    return UserAgent().random


def request(site: pywikibot.site.BaseSite,
            uri: str | None = None,
            headers: dict | None = None,
            **kwargs) -> requests.Response:
    """Request to Site with default error handling and response decoding.

    See :py:obj:`requests.Session.request` for additional parameters.

    The optional uri is a relative uri from site base uri including the
    document root '/'.

    .. versionchanged:: 8.2
       a *protocol* parameter can be given which is passed to the
       :meth:`family.Family.base_url` method.

    :param site: The Site to connect to
    :param uri: the URI to retrieve
    :keyword CodecInfo or str or None charset: Either a valid charset
        (usable for `str.decode()`) or None to automatically chose the
        charset from the returned header (defaults to latin-1).
    :keyword str | None protocol: a url scheme
    :return: The received data Response
    """
    kwargs.setdefault('verify', site.verify_SSL_certificate())
    if not headers:
        headers = {}
        format_string = None
    else:
        format_string = headers.get('user-agent')
    headers['user-agent'] = user_agent(site, format_string)

    baseuri = site.base_url(uri, protocol=kwargs.pop('protocol', None))
    r = fetch(baseuri, headers=headers, **kwargs)
    site.throttle.retry_after = int(r.headers.get('retry-after', 0))
    return r


def get_authentication(uri: str) -> tuple[str, str] | None:
    """Retrieve authentication token.

    :param uri: the URI to access
    :return: authentication token
    """
    parsed_uri = urlparse(uri)
    netloc_parts = parsed_uri.netloc.split('.')
    netlocs = [parsed_uri.netloc] + ['.'.join(['*'] + netloc_parts[i + 1:])
                                     for i in range(len(netloc_parts))]
    for path in netlocs:
        if path in config.authenticate:
            length = len(config.authenticate[path])
            if length in (2, 4):
                return config.authenticate[path]

            warn(f'config.authenticate[{path!r}] has invalid value.\n'
                 f'It should contain 2 or 4 items, not {length}.\n'
                 f'See {pywikibot.__url__}/OAuth for more info.', stacklevel=2)

    return None


def error_handling_callback(response) -> None:
    """Raise exceptions and log alerts.

    :param response: Response returned by Session.request().
    :type response: :py:obj:`requests.Response`
    """
    # TODO: do some error correcting stuff
    if isinstance(response, requests.exceptions.SSLError) \
       and 'certificate verify failed' in str(response):
        raise FatalServerError(str(response))

    if isinstance(response, requests.ConnectionError):
        msg = str(response)
        if ('NewConnectionError' in msg or 'NameResolutionError' in msg) \
           and re.search(r'\[Errno (-2|8|11001)\]', msg):
            raise ConnectionError(response)

    # catch requests.ReadTimeout and requests.ConnectTimeout and convert
    # it to ServerError
    if isinstance(response, requests.Timeout):
        raise ServerError(response)

    if isinstance(response, ValueError):
        # MissingSchema, InvalidSchema, InvalidURL, InvalidHeader
        raise FatalServerError(str(response))

    if isinstance(response, Exception):
        with suppress(Exception):
            # request exception may contain response and request attribute
            error('An error occurred for uri ' + response.request.url)
        raise response from None

    if response.status_code == HTTPStatus.REQUEST_URI_TOO_LONG:
        raise Client414Error(HTTPStatus(response.status_code).description)

    if response.status_code == HTTPStatus.GATEWAY_TIMEOUT:
        raise Server504Error(
            f'Server {urlparse(response.url).netloc} timed out')

    if (not response.ok
            and response.status_code >= HTTPStatus.INTERNAL_SERVER_ERROR):
        raise ServerError(
            f'{response.status_code} Server Error: {response.reason}')

    # TODO: shall it raise? this might break some code, TBC
    # response.raise_for_status()

    # HTTP status 207 is also a success status for Webdav FINDPROP,
    # used by the version module.
    if response.status_code not in (HTTPStatus.OK, HTTPStatus.MULTI_STATUS):
        warning(f'Http response status {response.status_code}')


def fetch(uri: str, method: str = 'GET', headers: dict | None = None,
          default_error_handling: bool = True,
          use_fake_user_agent: bool | str = False, **kwargs):
    """HTTP request.

    See :py:obj:`requests.Session.request` for parameters.

    :param uri: URL to send
    :param method: HTTP method of the request (default: GET)
    :param headers: dictionary of headers of the request
    :param default_error_handling: Use default error handling
    :param use_fake_user_agent: Set to True to use fake UA, False to use
        pywikibot's UA, str to specify own UA. This behaviour might be
        overridden by domain in config.

    :keyword charset: Either a valid charset (usable for str.decode()) or None
        to automatically chose the charset from the returned header (defaults
        to latin-1)
    :type charset: CodecInfo, str, None
    :keyword verify: verify the SSL certificate (default is True)
    :type verify: bool or path to certificates
    :keyword callbacks: Methods to call once data is fetched
    :type callbacks: list of callable
    :rtype: :py:obj:`requests.Response`
    """
    # Change user agent depending on fake UA settings.
    # Set header to new UA if needed.
    headers = headers or {}
    headers.update(config.extra_headers.copy() or {})

    def assign_fake_user_agent(use_fake_user_agent, uri):
        uri_domain = urlparse(uri).netloc
        use_fake_user_agent = config.fake_user_agent_exceptions.get(
            uri_domain, use_fake_user_agent)

        if use_fake_user_agent is False:
            return user_agent()
        if use_fake_user_agent is True:
            return fake_user_agent()
        if use_fake_user_agent and isinstance(use_fake_user_agent, str):
            return use_fake_user_agent  # Custom UA.
        raise ValueError('Invalid parameter: '
                         f'{use_fake_user_agent=!s}')

    def assign_user_agent(user_agent_format_string):
        if not user_agent_format_string or '{' in user_agent_format_string:
            return user_agent(None, user_agent_format_string)
        # do nothing, it is already a UA
        return user_agent_format_string

    # If not already specified.
    if 'user-agent' not in headers:
        # Get fake UA exceptions from `fake_user_agent_exceptions` config.
        headers['user-agent'] = assign_fake_user_agent(use_fake_user_agent,
                                                       uri)
    # Already specified.
    else:
        headers['user-agent'] = assign_user_agent(headers.get('user-agent'))

    callbacks = kwargs.pop('callbacks', [])
    # error_handling_callback will be executed first.
    if default_error_handling:
        callbacks.insert(0, error_handling_callback)

    charset = kwargs.pop('charset', None)

    auth = get_authentication(uri)
    if auth is not None and len(auth) == 4:
        if isinstance(requests_oauthlib, ImportError):
            raise ModuleNotFoundError(f"""{requests_oauthlib}. Install it with

    pip install requests_oauthlib
""")

        auth = requests_oauthlib.OAuth1(*auth)

    timeout = config.socket_timeout

    try:
        # Note that the connections are pooled which mean that a future
        # HTTPS request can succeed even if the certificate is invalid and
        # verify=True, when a request with verify=False happened before
        response = session.request(method, uri,
                                   headers=headers, auth=auth, timeout=timeout,
                                   **kwargs)
    except Exception as e:
        response = e
    else:
        response.encoding = _decide_encoding(response, charset)

    for callback in callbacks:
        callback(response)

    return response


# Extract charset (from content-type header)
CHARSET_RE = re.compile(
    r'charset\s*=\s*(?P<q>[\'"]?)(?P<charset>[^\'",;>/]+)(?P=q)',
    flags=re.IGNORECASE,
)


def get_charset_from_content_type(content_type: str) -> str | None:
    """Get charset from the content-type header.

    .. versionadded:: 7.3
    """
    m = CHARSET_RE.search(content_type)
    if not m:
        return None
    charset = m['charset'].strip('"\' ').lower()
    # Convert to python correct encoding names
    if re.sub(r'[ _\-]', '', charset) == 'xeucjp':
        charset = 'euc_jp'
    else:
        # fix cp encodings (T304830, T307760, T312230)
        # remove delimiter in front of the code number
        # replace win/windows with cp
        # remove language code in font of win/windows
        charset = re.sub(
            r'\A(?:cp[ _\-]|(?:[a-z]+[_\-]?)?win(?:dows)?[_\-]?)(\d{3,4})',
            r'cp\1', charset)
    return charset


def _get_encoding_from_response_headers(
    response: requests.Response
) -> str | None:
    """Return charset given by the response header."""
    content_type = response.headers.get('content-type')

    if not content_type:
        return None

    charset = get_charset_from_content_type(content_type)
    if charset:
        header_encoding = charset
    elif 'json' in content_type:
        # application/json | application/sparql-results+json
        header_encoding = 'utf-8'
    elif 'xml' in content_type:
        header = response.content[:100].splitlines()[0]  # bytes
        m = re.search(
            br'encoding=(["\'])(?P<encoding>.+?)\1', header)
        header_encoding = m['encoding'].decode('utf-8') if m else 'utf-8'
    else:
        header_encoding = None

    return header_encoding


def _decide_encoding(response: requests.Response,
                     charset: str | None = None) -> str | None:
    """Detect the response encoding.

    .. versionchanged:: 10.1
       retrieve charset from `Accept-Charset` list which may look like
       `'ISO-8859-1,utf-8;q=0.7,*;q=0.7'`.

    :meta public:
    """
    def _try_decode(content: bytes, encoding: str | None) -> str | None:
        """Helper function to try decoding."""
        if encoding is None:
            return None

        try:
            content.decode(encoding)
        except LookupError:
            pywikibot.warning(
                f'Unknown or invalid encoding {encoding!r} for {response.url}')
        except UnicodeDecodeError as e:
            pywikibot.warning(f'{e} found in {content!r}')
        else:
            return encoding

        return None  # let chardet do the job

    header_encoding = _get_encoding_from_response_headers(response)
    if header_encoding is None:
        pywikibot.log('Http response does not contain a charset.')

    if charset is None:
        charset = response.request.headers.get('accept-charset', '')
        charset = charset.split(',', 1)[0].split(';', 1)[0].strip() or None

    # No charset requested, or in request headers or response headers.
    # Defaults to latin1.
    if charset is None:
        return _try_decode(response.content, header_encoding or 'latin1')

    if header_encoding is None:
        return _try_decode(response.content, charset)

    # Both charset and header_encoding are available.
    header_codecs = charset_codecs = None
    with suppress(LookupError):
        header_codecs = codecs.lookup(header_encoding)

    with suppress(LookupError):
        charset_codecs = codecs.lookup(charset)

    if header_codecs and charset_codecs and header_codecs != charset_codecs:
        pywikibot.warning(
            f'Encoding "{charset}" requested but "{header_encoding}" received'
            ' in the response header.'
        )

    return _try_decode(response.content, header_encoding) \
        or _try_decode(response.content, charset)



"""Wikibase data type classes."""
#
# (C) Pywikibot team, 2013-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import abc
import datetime
import json
import math
import re
from collections.abc import Mapping
from contextlib import suppress
from decimal import Decimal
from typing import TYPE_CHECKING, Any, cast

import pywikibot
from pywikibot import exceptions
from pywikibot.backports import Iterator
from pywikibot.time import Timestamp
from pywikibot.tools import (
    deprecated_signature,
    issue_deprecation_warning,
    remove_last_args,
)


if TYPE_CHECKING:
    from typing import Union

    from pywikibot.site import APISite, BaseSite, DataSite

    ItemPageStrNoneType = Union[str, pywikibot.ItemPage, None]
    ToDecimalType = Union[int, float, str, Decimal, None]


__all__ = (
    'Coordinate',
    'WbGeoShape',
    'WbMonolingualText',
    'WbQuantity',
    'WbTabularData',
    'WbTime',
    'WbUnknown',
)


class WbRepresentation(abc.ABC):

    """Abstract class for Wikibase representations."""

    _items: tuple[str, ...]

    @abc.abstractmethod
    def __init__(self) -> None:
        """Initializer."""
        raise NotImplementedError

    @abc.abstractmethod
    def toWikibase(self) -> Any:
        """Convert representation to JSON for the Wikibase API."""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def fromWikibase(
        cls,
        data: dict[str, Any],
        site: DataSite | None = None,
    ) -> WbRepresentation:
        """Create a representation object based on JSON from Wikibase API."""
        raise NotImplementedError

    def __str__(self) -> str:
        return json.dumps(
            self.toWikibase(),
            indent=4,
            sort_keys=True,
            separators=(',', ': ')
        )

    def __repr__(self) -> str:
        """String representation of this object.

        .. versionchanged:: 10.4
           Parameters are shown as representations instead of plain
           strings.

        :meta public:
        """
        assert isinstance(self._items, tuple)
        assert all(isinstance(item, str) for item in self._items)

        attrs = ', '.join(f'{attr}={getattr(self, attr)!r}'
                          for attr in self._items)
        return f'{type(self).__name__}({attrs})'

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self.toWikibase() == other.toWikibase()

        return NotImplemented

    def __hash__(self) -> int:
        return hash(json.dumps(self.toWikibase(), sort_keys=True))


class Coordinate(WbRepresentation):

    """Class for handling and storing Coordinates."""

    _items = ('lat', 'lon', 'entity')

    @deprecated_signature(since='10.4.0')
    def __init__(
        self,
        lat: float,
        lon: float,
        *,
        alt: float | None = None,
        precision: float | None = None,
        globe: str | None = None,
        typ: str = '',
        name: str = '',
        dim: int | None = None,
        site: DataSite | None = None,
        globe_item: ItemPageStrNoneType = None,
        primary: bool = False
    ) -> None:
        """Represent a geo coordinate.

        .. versionchanged:: 10.4
           The parameters after `lat` and `lon` are now keyword-only.

        :param lat: Latitude coordinate
        :param lon: Longitude coordinate
        :param alt: Altitude in meters
        :param precision: Precision of the coordinate
        :param globe: The globe the coordinate is on (e.g. 'earth')
        :param typ: Type of coordinate point
        :param name: Name associated with the coordinate
        :param dim: Dimension in meters used for precision calculation
        :param site: The Wikibase site instance
        :param globe_item: Wikibase item or entity URI for the globe;
            takes precedence over *globe*
        :param primary: Indicates if this is a primary coordinate set
            (default: False)
        """
        self.lat = lat
        self.lon = lon
        self.alt = alt
        self._precision = precision
        self._entity = globe_item
        self.type = typ
        self.name = name
        self._dim = dim
        self.site = site or pywikibot.Site().data_repository()
        self.primary = primary

        if globe:
            globe = globe.lower()
        elif not globe_item:
            globe = self.site.default_globe()
        self.globe = globe

    @property
    def entity(self) -> str:
        """Return the entity URI of the globe.

        :raises CoordinateGlobeUnknownError: the globe is not supported
            by Wikibase
        """
        if not self._entity:
            if self.globe not in self.site.globes():
                raise exceptions.CoordinateGlobeUnknownError(
                    f'{self.globe} is not supported in Wikibase yet.')

            return self.site.globes()[self.globe]

        if isinstance(self._entity, pywikibot.ItemPage):
            return self._entity.concept_uri()

        return self._entity

    def toWikibase(self) -> dict[str, Any]:
        """Export the data to a JSON object for the Wikibase API.

        :return: Wikibase JSON representation of the coordinate
        """
        return {
            'latitude': self.lat,
            'longitude': self.lon,
            'altitude': self.alt,
            'globe': self.entity,
            'precision': self.precision,
        }

    @classmethod
    def fromWikibase(cls, data: dict[str, Any],
                     site: DataSite | None = None) -> Coordinate:
        """Create an object from Wikibase's JSON output.

        :param data: Wikibase JSON data
        :param site: The Wikibase site instance
        :return: Coordinate instance
        """
        site = site or pywikibot.Site().data_repository()
        globe = None

        if data.get('globe'):
            globes = {entity: name for name, entity in site.globes().items()}
            globe = globes.get(data['globe'])

        return cls(
            data['latitude'],
            data['longitude'],
            alt=data.get('altitude'),
            precision=data.get('precision'),
            globe=globe,
            site=site,
            globe_item=data.get('globe')
        )

    @property
    def precision(self) -> float | None:
        """Return the precision of the geo coordinate.

        The precision is calculated if the Coordinate does not have a
        precision, and self._dim is set.

        When no precision and no self._dim exists, None is returned.

        The biggest error (in degrees) will be given by the longitudinal
        error; the same error in meters becomes larger (in degrees)
        further up north. We can thus ignore the latitudinal error.

        The longitudinal can be derived as follows:

        In small angle approximation (and thus in radians):

        :math:`M{Δλ ≈ Δpos / r_φ}`, where :math:`r_φ` is the radius of
        earth at the given latitude. :math:`Δλ` is the error in
        longitude.

        :math:`M{r_φ = r cos(φ)}`, where :math:`r` is the radius of
        earth, :math:`φ` the latitude

        Therefore:

        .. code-block:: python

           precision = math.degrees(
               self._dim / (radius * math.cos(math.radians(self.lat))))

        :return: precision in degrees or None
        """
        if self._precision is not None:
            return self._precision

        if self._dim is None:
            return None

        radius = 6378137  # Earth radius in meters (TODO: support other globes)
        with suppress(ZeroDivisionError):
            self._precision = math.degrees(
                self._dim / (radius * math.cos(math.radians(self.lat))))

        return self._precision

    @precision.setter
    def precision(self, value: float) -> None:
        """Set the precision value.

        :param value: precision in degrees
        """
        self._precision = value

    def precisionToDim(self) -> int | None:
        """Convert precision from Wikibase to GeoData's dim.

        dim is calculated if the Coordinate doesn't have a dimension, and
        precision is set. When neither dim nor precision are set, ValueError
        is thrown.

        Carrying on from the earlier derivation of precision, since

        .. code-block:: python

           precision = math.degrees(
               dim / (radius * math.cos(math.radians(self.lat))))

        we get:

        .. code-block:: python

           dim = math.radians(
               precision) * radius * math.cos(math.radians(self.lat))

        But this is not valid, since it returns a float value for dim which is
        an integer. We must round it off to the nearest integer.

        Therefore:

        .. code-block:: python

           dim = int(round(math.radians(
               precision)*radius*math.cos(math.radians(self.lat))))

        :return: dimension in meters
        :raises ValueError: if neither dim nor precision is set
        """
        if self._dim is not None:
            return self._dim

        if self._precision is None:
            raise ValueError('No values set for dim or precision')

        radius = 6378137
        self._dim = int(
            round(
                math.radians(self._precision) * radius * math.cos(
                    math.radians(self.lat))
            )
        )
        return self._dim

    @deprecated_signature(since='10.4.0')
    def get_globe_item(self, repo: DataSite | None = None, *,
                       lazy_load: bool = False) -> pywikibot.ItemPage:
        """Return the ItemPage corresponding to the globe.

        .. note:: The globe need not be in the same data repository as
           the Coordinate itself.

        A successful lookup is stored as an internal value to avoid the
        need for repeated lookups.

        .. versionchanged:: 10.4
           The *lazy_load* parameter is now keyword-only.

        :param repo: the Wikibase site for the globe, if different from
            that provided with the Coordinate
        :param lazy_load: Do not raise :exc:`exceptions.NoPageError` if
            ItemPage does not exist
        :return: pywikibot.ItemPage of the globe
        """
        if isinstance(self._entity, pywikibot.ItemPage):
            return self._entity

        repo = repo or self.site
        return pywikibot.ItemPage.from_entity_uri(repo, self.entity, lazy_load)


class _Precision(Mapping):

    """Wrapper for WbTime.PRECISION to deprecate 'millenia' key."""

    PRECISION = {
        '1000000000': 0,
        '100000000': 1,
        '10000000': 2,
        '1000000': 3,
        '100000': 4,
        '10000': 5,
        'millennium': 6,
        'century': 7,
        'decade': 8,
        'year': 9,
        'month': 10,
        'day': 11,
        'hour': 12,
        'minute': 13,
        'second': 14,
    }

    def __getitem__(self, key: str) -> int:
        if key == 'millenia':
            issue_deprecation_warning(
                f'{key!r} key for precision', "'millennium'", since='10.0.0')
            return self.PRECISION['millennium']

        return self.PRECISION[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self.PRECISION)

    def __len__(self) -> int:
        return len(self.PRECISION)


class WbTime(WbRepresentation):

    """A Wikibase time representation.

    Make a WbTime object from the current time:

    .. code-block:: python

        current_ts = pywikibot.Timestamp.now()
        wbtime = pywikibot.WbTime.fromTimestamp(current_ts)

    For converting python datetime objects to WbTime objects, see
    :class:`pywikibot.Timestamp` and :meth:`fromTimestamp`.
    """

    month: int
    day: int
    hour: int
    minute: int
    second: int

    PRECISION = _Precision()

    FORMATSTR = '{0:+012d}-{1:02d}-{2:02d}T{3:02d}:{4:02d}:{5:02d}Z'

    _items = ('year', 'month', 'day', 'hour', 'minute', 'second',
              'precision', 'before', 'after', 'timezone', 'calendarmodel')

    _month_offset = {
        1: 0,
        2: 31,  # Jan -> Feb: 31 days
        3: 59,  # Feb -> Mar: 28 days, plus 31 days in Jan -> Feb
        4: 90,  # Mar -> Apr: 31 days, plus 59 days in Jan -> Mar
        5: 120,  # Apr -> May: 30 days, plus 90 days in Jan -> Apr
        6: 151,  # May -> Jun: 31 days, plus 120 days in Jan -> May
        7: 181,  # Jun -> Jul: 30 days, plus 151 days in Jan -> Jun
        8: 212,  # Jul -> Aug: 31 days, plus 181 days in Jan -> Jul
        9: 243,  # Aug -> Sep: 31 days, plus 212 days in Jan -> Aug
        10: 273,  # Sep -> Oct: 30 days, plus 243 days in Jan -> Sep
        11: 304,  # Oct -> Nov: 31 days, plus 273 days in Jan -> Oct
        12: 334,  # Nov -> Dec: 30 days, plus 304 days in Jan -> Nov
    }

    _timestr_re = re.compile(
        r'([-+]?\d{1,16})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2})Z')

    @deprecated_signature(since='10.4.0')
    def __init__(
        self,
        year: int,
        month: int | None = None,
        day: int | None = None,
        hour: int | None = None,
        minute: int | None = None,
        second: int | None = None,
        *,
        precision: int | str | None = None,
        before: int = 0,
        after: int = 0,
        timezone: int = 0,
        calendarmodel: str | None = None,
        site: DataSite | None = None
    ) -> None:
        """Create a new WbTime object.

        The precision can be set by the Wikibase int value (0-14) or by
        a human readable string, e.g., ``hour``. If no precision is
        given, it is set according to the given time units.

        Timezone information is given in three different ways depending
        on the time:

        * Times after the implementation of UTC (1972): as an offset
          from UTC in minutes;
        * Times before the implementation of UTC: the offset of the time
          zone from universal time;
        * Before the implementation of time zones: The longitude of the
          place of the event, in the range −180° to 180°, multiplied by
          4 to convert to minutes.

        .. note::  **Comparison information:** When using the greater
           than or equal to operator, or the less than or equal to
           operator, two different time objects with the same UTC time
           after factoring in timezones are considered equal. However,
           the equality operator will return false if the timezone is
           different.

        .. deprecated:: 10.0
           *precision* value 'millenia' is deprecated; 'millennium' must
           be used instead.

        .. versionchanged:: 10.4
           The parameters except timestamp values are now keyword-only.
           A TypeError is raised if *year* is not an int. Previously, a
           ValueError was raised if *year* was None.

        :param year: The year as a signed integer of between 1 and 16
            digits.
        :param month: Month of the timestamp, if it exists.
        :param day: Day of the timestamp, if it exists.
        :param hour: Hour of the timestamp, if it exists.
        :param minute: Minute of the timestamp, if it exists.
        :param second: Second of the timestamp, if it exists.
        :param precision: The unit of the precision of the time. Must be
            either an int in range 0 - 14 or one of '1000000000',
            '100000000', '10000000', '1000000', '100000', '10000',
            'millennium', 'century', 'decade', 'year', month', 'day',
            'hour', 'minute' or 'second'.
        :param before: Number of units after the given time it could be,
            if uncertain. The unit is given by the precision.
        :param after: Number of units before the given time it could be,
            if uncertain. The unit is given by the precision.
        :param timezone: Timezone information in minutes.
        :param calendarmodel: URI identifying the calendar model.
        :param site: The Wikibase site. If not provided, retrieves the
            data repository from the default site from user-config.py.
            Only used if calendarmodel is not given.
        :raises TypeError: Invalid *year* type.
        :raises ValueError: Invalid *precision* or *site* or default
            site has no data repository.
        """
        if not isinstance(year, int):
            raise TypeError(f'year must be an int, not {type(year).__name__}')

        units = [
            ('month', month, 1),
            ('day', day, 1),
            ('hour', hour, 0),
            ('minute', minute, 0),
            ('second', second, 0),
        ]

        # set unit attribute values
        self.year = year
        for unit, value, default in units:
            setattr(self, unit, value if value is not None else default)

        if precision is None:
            # Autodetection of precision based on the passed time values
            prec = self.PRECISION['year']

            for unit, value, _ in units:
                if value is not None:
                    prec = self.PRECISION[unit]
        else:
            # explicit precision is given
            if (isinstance(precision, int)
                    and precision in self.PRECISION.values()):
                prec = precision
            elif isinstance(precision, str) and precision in self.PRECISION:
                prec = self.PRECISION[cast(str, precision)]
            else:
                raise ValueError(f'Invalid precision: "{precision}"')

        self.precision = prec
        self.after = after
        self.before = before
        self.timezone = timezone
        if calendarmodel is None:
            site = site or pywikibot.Site().data_repository()
            if site is None:
                raise ValueError(
                    f'Site {pywikibot.Site()} has no data repository')
            calendarmodel = site.calendarmodel()

        self.calendarmodel = calendarmodel

    def _getSecondsAdjusted(self) -> int:
        """Return an internal representation of the time object as seconds.

        The value adjusts itself for timezones. It is not compatible
        with before/after.

        This value should *only* be used for comparisons, and its value
        may change without warning.

        .. versionadded:: 8.0

        :return: An integer roughly representing the number of seconds
            since January 1, 0000 AD, adjusted for leap years.
        """
        # This function ignores leap seconds. Since it is not required
        # to correlate to an actual UNIX timestamp, this is acceptable.

        # We are always required to have a year.
        elapsed_seconds = int(self.year * 365.25 * 24 * 60 * 60)
        if self.month > 1:
            elapsed_seconds += self._month_offset[self.month] * 24 * 60 * 60
            # The greogrian calendar
            if (self.calendarmodel == 'http://www.wikidata.org/entity/Q1985727'
                and (self.year % 400 == 0
                     or (self.year % 4 == 0 and self.year % 100 != 0)
                     and self.month > 2)):
                elapsed_seconds += 24 * 60 * 60  # Leap year
            # The julian calendar
            if (self.calendarmodel == 'http://www.wikidata.org/entity/Q1985786'
                    and self.year % 4 == 0 and self.month > 2):
                elapsed_seconds += 24 * 60 * 60
        if self.day > 1:
            # Days start at 1, not 0.
            elapsed_seconds += (self.day - 1) * 24 * 60 * 60
        elapsed_seconds += self.hour * 60 * 60
        elapsed_seconds += self.minute * 60
        elapsed_seconds += self.second
        if self.timezone is not None:
            # See T325866
            elapsed_seconds -= self.timezone * 60
        return elapsed_seconds

    def __lt__(self, other: object) -> bool:
        """Compare if self is less than other.

        .. versionadded:: 8.0
        """
        if isinstance(other, WbTime):
            return self._getSecondsAdjusted() < other._getSecondsAdjusted()
        return NotImplemented

    def __le__(self, other: object) -> bool:
        """Compare if self is less equals other.

        .. versionadded:: 8.0
        """
        if isinstance(other, WbTime):
            return self._getSecondsAdjusted() <= other._getSecondsAdjusted()
        return NotImplemented

    def __gt__(self, other: object) -> bool:
        """Compare if self is greater than other.

        .. versionadded:: 8.0
        """
        if isinstance(other, WbTime):
            return self._getSecondsAdjusted() > other._getSecondsAdjusted()
        return NotImplemented

    def __ge__(self, other: object) -> bool:
        """Compare if self is greater equals other.

        .. versionadded:: 8.0
        """
        if isinstance(other, WbTime):
            return self._getSecondsAdjusted() >= other._getSecondsAdjusted()
        return NotImplemented

    def equal_instant(self, other: WbTime) -> bool:
        """Checks if the two times represent the same instant in time.

        This is different from the equality operator, which will return false
        for two times that are the same number of UTC seconds, but with
        different timezone information.

        For example, a time with at 10:00 UTC-5 would return false if checked
        with == with a time at 15:00 UTC, but would return true with
        this method.

        .. versionadded:: 9.0
        """
        return self._getSecondsAdjusted() == other._getSecondsAdjusted()

    @classmethod
    @deprecated_signature(since='10.4.0')
    def fromTimestr(
        cls,
        datetimestr: str,
        *,
        precision: int | str = 14,
        before: int = 0,
        after: int = 0,
        timezone: int = 0,
        calendarmodel: str | None = None,
        site: DataSite | None = None
    ) -> WbTime:
        """Create a new WbTime object from a UTC date/time string.

        The timestamp format must match a string resembling ISO 8601
        with the following constraints:

        - Year is signed and can have between 1 and 16 digits.
        - Month, day, hour, minute and second are always two digits.
          They may be zero.
        - Time is always in UTC and ends with ``Z``.
        - Example: ``+0000000000123456-01-01T00:00:00Z``.

        .. versionchanged:: 10.4
           The parameters except *datetimestr* are now keyword-only.

        :param datetimestr: Timestamp string to parse
        :param precision: The unit of the precision of the time. Defaults
            to 14 (second).
        :param before: Number of units after the given time it could be,
            if uncertain. The unit is given by the precision.
        :param after: Number of units before the given time it could be,
            if uncertain. The unit is given by the precision.
        :param timezone: Timezone offset in minutes.
        :param calendarmodel: URI identifying the calendar model.
        :param site: The Wikibase site. If not provided, retrieves the
            data repository from the default site from user-config.py.
            Only used if calendarmodel is not given.
        :raises ValueError: If the string does not match the expected
            format.
        """
        match = cls._timestr_re.match(datetimestr)
        if not match:
            raise ValueError(f"Invalid format: '{datetimestr}'")

        t = match.groups()
        return cls(int(t[0]), int(t[1]), int(t[2]),
                   int(t[3]), int(t[4]), int(t[5]),
                   precision=precision, before=before, after=after,
                   timezone=timezone, calendarmodel=calendarmodel, site=site)

    @classmethod
    @deprecated_signature(since='10.4.0')
    def fromTimestamp(
        cls,
        timestamp: Timestamp,
        *,
        precision: int | str = 14,
        before: int = 0,
        after: int = 0,
        timezone: int = 0,
        calendarmodel: str | None = None,
        site: DataSite | None = None,
        copy_timezone: bool = False
    ) -> WbTime:
        """Create a new WbTime object from a pywikibot.Timestamp.

        .. versionchanged:: 8.0
           Added *copy_timezone* parameter.
        .. versionchanged:: 10.4
           The parameters except *timestamp* are now keyword-only.


        :param timestamp: Timestamp
        :param precision: The unit of the precision of the time.
            Defaults to 14 (second).
        :param before: Number of units after the given time it could be,
            if uncertain. The unit is given by the precision.
        :param after: Number of units before the given time it could be,
            if uncertain. The unit is given by the precision.
        :param timezone: Timezone information in minutes.
        :param calendarmodel: URI identifying the calendar model.
        :param site: The Wikibase site. If not provided, retrieves the
            data repository from the default site from user-config.py.
            Only used if calendarmodel is not given.
        :param copy_timezone: Whether to copy the timezone from the
            timestamp if it has timezone information. Defaults to False
            to maintain backwards compatibility. If a timezone is given,
            timezone information is discarded.
        """
        if not timezone and timestamp.tzinfo and copy_timezone:
            timezone = int(timestamp.utcoffset().total_seconds() / 60)
        return cls.fromTimestr(timestamp.isoformat(), precision=precision,
                               before=before, after=after, timezone=timezone,
                               calendarmodel=calendarmodel, site=site)

    @staticmethod
    def _normalize_millennium(year: int) -> int:
        """Round the given year to the start of its millennium.

        The rounding is performed towards positive infinity for positive
        years and towards negative infinity for negative years.

        .. versionadded:: 10.4

        :param year: The year as an integer.
        :return: The first year of the millennium containing the given
            year.
        """
        # For negative years, floor rounds away from zero to correctly handle
        # BCE dates. For positive years, ceil rounds up to the next
        # millennium/century.
        year_float = year / 1000
        if year_float < 0:
            year = math.floor(year_float)
        else:
            year = math.ceil(year_float)
        return year * 1000

    @staticmethod
    def _normalize_century(year: int) -> int:
        """Round the given year to the start of its century.

        The rounding is performed towards positive infinity for positive
        years and towards negative infinity for negative years.

        .. versionadded:: 10.4

        :param year: The year as an integer.
        :return: The first year of the century containing the given year.
        """
        # For century, -1301 is the same century as -1400 but not -1401.
        # Similar for 1901 and 2000 vs 2001.
        year_float = year / 100
        if year_float < 0:
            year = math.floor(year_float)
        else:
            year = math.ceil(year_float)
        return year * 100

    @staticmethod
    def _normalize_decade(year: int) -> int:
        """Round the given year down to the start of its decade.

        Unlike millennium or century normalization, this always
        truncates towards zero.

        .. versionadded:: 10.4

        :param year: The year as an integer.
        :return: The first year of the decade containing the given year.
        """
        # For decade, -1340 is the same decade as -1349 but not -1350.
        # Similar for 2010 and 2019 vs 2020
        year_float = year / 10
        year = math.trunc(year_float)
        return year * 10

    @staticmethod
    def _normalize_power_of_ten(year: int, precision: int) -> int:
        """Round the year to the given power-of-ten precision.

        This is used for very coarse historical precision levels, where
        the time unit represents a power-of-ten number of years.

        .. versionadded:: 10.4

        :param year: The year as an integer.
        :param precision: The precision level (Wikibase int value).
        :return: The normalized year rounded to the nearest matching
            power-of-ten boundary.
        """
        # Wikidata rounds the number based on the first non-decimal digit.
        # Python's round function will round -15.5 to -16, and +15.5 to +16
        # so we don't need to do anything complicated like the other
        # examples.
        power_of_10 = 10 ** (9 - precision)
        return round(year / power_of_10) * power_of_10

    def normalize(self) -> WbTime:
        """Normalizes the WbTime object to account for precision.

        Normalization is needed because WbTime objects can have hidden
        values that affect naive comparisons, such as an object set to
        a precision of YEAR but containing a month and day value.

        This function returns a new normalized object and does not do
        any modification in place.

        Normalization will delete timezone information if the precision
        is less than or equal to DAY.

        .. note:: Normalized WbTime objects can only be compared to
           other normalized WbTime objects of the same precision.
           Normalization might make a WbTime object that was less than
           another WbTime object before normalization, greater than it
           after normalization, or vice versa.
        """
        year = self.year
        for prec in 'millennium', 'century', 'decade':
            if self.precision == self.PRECISION[prec]:
                handler = getattr(self, '_normalize_' + prec)
                year = handler(year)
                break
        else:
            lower = self.PRECISION['1000000000']
            upper = self.PRECISION['10000']
            if lower <= self.precision <= upper:
                year = self._normalize_power_of_ten(year, self.precision)

        kwargs = {
            'precision': self.precision,
            'before': self.before,
            'after': self.after,
            'calendarmodel': self.calendarmodel,
            'year': year
        }

        for prec in 'month', 'day', 'hour', 'minute', 'second':
            if self.precision >= self.PRECISION[prec]:
                kwargs[prec] = getattr(self, prec)
                if prec == 'hour':
                    # Add timezone, see T326693
                    kwargs['timezone'] = self.timezone

        return type(self)(**kwargs)

    @remove_last_args(['normalize'])  # since 8.2.0
    def toTimestr(self, force_iso: bool = False) -> str:
        """Convert the data to a UTC date/time string.

        .. seealso:: :meth:`fromTimestr` for differences between output
           with and without *force_iso* parameter.

        .. versionchanged:: 8.0
           *normalize* parameter was added.
        .. versionchanged:: 8.2
           *normalize* parameter was removed due to :phab:`T340495` and
           :phab:`T57755`

        :param force_iso: whether the output should be forced to ISO 8601
        :return: Timestamp in a format resembling ISO 8601
        """
        if force_iso:
            return Timestamp._ISO8601Format_new.format(
                self.year, max(1, self.month), max(1, self.day),
                self.hour, self.minute, self.second)
        return self.FORMATSTR.format(self.year, self.month, self.day,
                                     self.hour, self.minute, self.second)

    def toTimestamp(self, timezone_aware: bool = False) -> Timestamp:
        """Convert the data to a pywikibot.Timestamp.

        .. versionchanged:: 8.0.1
           *timezone_aware* parameter was added.

        :param timezone_aware: Whether the timezone should be passed to
            the Timestamp object.
        :raises ValueError: instance value cannot be represented using
            Timestamp
        """
        if self.year <= 0:
            raise ValueError('You cannot turn BC dates into a Timestamp')
        ts = Timestamp.fromISOformat(
            self.toTimestr(force_iso=True).lstrip('+'))
        if timezone_aware:
            ts = ts.replace(tzinfo=datetime.timezone(
                datetime.timedelta(minutes=self.timezone)))
        return ts

    @remove_last_args(['normalize'])  # since 8.2.0
    def toWikibase(self) -> dict[str, Any]:
        """Convert the data to a JSON object for the Wikibase API.

        .. versionchanged:: 8.0
           *normalize* parameter was added.
        .. versionchanged:: 8.2
           *normalize* parameter was removed due to :phab:`T340495` and
           :phab:`T57755`

        :return: Wikibase JSON
        """
        return {
            'time': self.toTimestr(),
            'precision': self.precision,
            'after': self.after,
            'before': self.before,
            'timezone': self.timezone,
            'calendarmodel': self.calendarmodel,
        }

    @classmethod
    def fromWikibase(cls, data: dict[str, Any],
                     site: DataSite | None = None) -> WbTime:
        """Create a WbTime from the JSON data given by the Wikibase API.

        :param data: Wikibase JSON
        :param site: The Wikibase site. If not provided, retrieves the
            data repository from the default site from user-config.py.
        """
        return cls.fromTimestr(
            data['time'],
            precision=data['precision'],
            before=data['before'],
            after=data['after'],
            timezone=data['timezone'],
            calendarmodel=data['calendarmodel'],
            site=site
        )


class WbQuantity(WbRepresentation):

    """A Wikibase quantity representation."""

    _items = ('amount', 'upperBound', 'lowerBound', 'unit')

    @staticmethod
    def _todecimal(value: ToDecimalType) -> Decimal | None:
        """Convert a string to a Decimal for use in WbQuantity.

        None value is returned as is.

        :param value: decimal number to convert
        """
        if isinstance(value, Decimal):
            return value
        if value is None:
            return None
        return Decimal(str(value))

    @staticmethod
    def _fromdecimal(value: Decimal | None) -> str | None:
        """Convert a Decimal to a string representation suitable for WikiBase.

        None value is returned as is.

        :param value: decimal number to convert
        """
        return format(value, '+g') if value is not None else None

    def __init__(
        self, amount: ToDecimalType,
        unit: ItemPageStrNoneType = None,
        error: ToDecimalType | tuple[ToDecimalType, ToDecimalType] = None,
        site: DataSite | None = None,
    ) -> None:
        """Create a new WbQuantity object.

        :param amount: number representing this quantity
        :param unit: the Wikibase item for the unit or the entity URI of
            this Wikibase item.
        :param error: the uncertainty of the amount (e.g. ±1)
        :param site: The Wikibase site
        """
        if amount is None:
            raise ValueError('no amount given')

        self.amount = self._todecimal(amount)
        self._unit = unit
        self.site = site or pywikibot.Site().data_repository()

        # also allow entity URIs to be provided via unit parameter
        if isinstance(unit, str) \
           and not unit.startswith(('http://', 'https://')):
            raise ValueError("'unit' must be an ItemPage or entity uri.")

        if error is None:
            self.upperBound = self.lowerBound = None
        else:
            if isinstance(error, tuple):
                upper_error = self._todecimal(error[0])
                lower_error = self._todecimal(error[1])
            else:
                upper_error = lower_error = self._todecimal(error)

            assert upper_error is not None and lower_error is not None
            assert self.amount is not None

            self.upperBound = self.amount + upper_error
            self.lowerBound = self.amount - lower_error

    @property
    def unit(self) -> str:
        """Return _unit's entity uri or '1' if _unit is None."""
        if isinstance(self._unit, pywikibot.ItemPage):
            return self._unit.concept_uri()
        return self._unit or '1'

    def get_unit_item(self, repo: DataSite | None = None,
                      lazy_load: bool = False) -> pywikibot.ItemPage:
        """Return the ItemPage corresponding to the unit.

        Note that the unit need not be in the same data repository as
        the WbQuantity itself.

        A successful lookup is stored as an internal value to avoid the
        need for repeated lookups.

        :param repo: the Wikibase site for the unit, if different from
            that provided with the WbQuantity.
        :param lazy_load: Do not raise NoPage if ItemPage does not
            exist.
        :return: pywikibot.ItemPage
        """
        if not isinstance(self._unit, str):
            return self._unit

        repo = repo or self.site
        self._unit = pywikibot.ItemPage.from_entity_uri(
            repo, self._unit, lazy_load)
        return self._unit

    def toWikibase(self) -> dict[str, Any]:
        """Convert the data to a JSON object for the Wikibase API.

        :return: Wikibase JSON
        """
        return {
            'amount': self._fromdecimal(self.amount),
            'upperBound': self._fromdecimal(self.upperBound),
            'lowerBound': self._fromdecimal(self.lowerBound),
            'unit': self.unit,
        }

    @classmethod
    def fromWikibase(cls, data: dict[str, Any],
                     site: DataSite | None = None) -> WbQuantity:
        """Create a WbQuantity from the JSON data given by the Wikibase API.

        :param data: Wikibase JSON
        :param site: The Wikibase site
        """
        amount = cls._todecimal(data['amount'])
        upper_bound = cls._todecimal(data.get('upperBound'))
        lower_bound = cls._todecimal(data.get('lowerBound'))
        bounds_provided = (upper_bound is not None and lower_bound is not None)
        error = None
        if bounds_provided:
            error = (upper_bound - amount, amount - lower_bound)
        unit = None if data['unit'] == '1' else data['unit']
        return cls(amount, unit, error, site)


class WbMonolingualText(WbRepresentation):

    """A Wikibase monolingual text representation."""

    _items = ('text', 'language')

    def __init__(self, text: str, language: str) -> None:
        """Create a new WbMonolingualText object.

        :param text: text string
        :param language: language code of the string
        """
        if not text or not language:
            raise ValueError('text and language cannot be empty')
        self.text = text
        self.language = language

    def toWikibase(self) -> dict[str, Any]:
        """Convert the data to a JSON object for the Wikibase API.

        :return: Wikibase JSON
        """
        return {
            'text': self.text,
            'language': self.language,
        }

    @classmethod
    def fromWikibase(cls, data: dict[str, Any],
                     site: DataSite | None = None) -> WbMonolingualText:
        """Create a WbMonolingualText from the JSON data given by Wikibase API.

        :param data: Wikibase JSON
        :param site: The Wikibase site
        """
        return cls(data['text'], data['language'])


class WbDataPage(WbRepresentation):

    """An abstract Wikibase representation for data pages.

    .. warning:: Perhaps a temporary implementation until :phab:`T162336`
       has been resolved.
    .. note:: that this class cannot be used directly.
    """

    _items = ('page', )

    @classmethod
    @abc.abstractmethod
    def _get_data_site(cls, repo_site: DataSite) -> APISite:
        """Return the site serving as a repository for a given data type.

        .. note:: implemented in the extended class.

        :param repo_site: The Wikibase site
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def _get_type_specifics(cls, site: DataSite) -> dict[str, Any]:
        """Return the specifics for a given data type.

        .. note:: Must be implemented in the extended class.

        The dict should have three keys:

        * ending: str, required filetype-like ending in page titles.
        * label: str, describing the data type for use in error messages.
        * data_site: APISite, site serving as a repository for
          the given data type.

        :param site: The Wikibase site
        """
        raise NotImplementedError

    @staticmethod
    def _validate(page: pywikibot.Page, data_site: BaseSite, ending: str,
                  label: str) -> None:
        """Validate the provided page against general and type specific rules.

        :param page: Page containing the data.
        :param data_site: The site serving as a repository for the given
            data type.
        :param ending: Required filetype-like ending in page titles.
            E.g. '.map'
        :param label: Label describing the data type in error messages.
        """
        if not isinstance(page, pywikibot.Page):
            raise ValueError(f'Page {page} must be a pywikibot.Page object '
                             f'not a {type(page)}.')

        # validate page exists
        if not page.exists():
            raise ValueError(f'Page {page} must exist.')

        # validate page is on the right site, and that site supports the type
        if not data_site:
            raise ValueError(
                f'The provided site does not support {label}.')
        if page.site != data_site:
            raise ValueError(
                f'Page must be on the {label} repository site.')

        # validate page title fulfills hard-coded Wikibase requirement
        # pcre regexp: '/^Data:[^\\[\\]#\\\:{|}]+\.map$/u' for geo-shape
        # pcre regexp: '/^Data:[^\\[\\]#\\\:{|}]+\.tab$/u' for tabular-data
        # As we have already checked for existence the following simplified
        # check should be enough.
        if not page.title().startswith('Data:') \
           or not page.title().endswith(ending):
            raise ValueError(f"Page must be in 'Data:' namespace and end in "
                             f"'{ending}' for {label}.")

    def __init__(self,
                 page: pywikibot.Page,
                 site: DataSite | None = None) -> None:
        """Create a new WbDataPage object.

        :param page: page containing the data
        :param site: The Wikibase site
        """
        site = site or page.site.data_repository()
        specifics = type(self)._get_type_specifics(site)
        WbDataPage._validate(page, specifics['data_site'], specifics['ending'],
                             specifics['label'])
        self.page = page

    def __hash__(self) -> int:
        """Override super.hash() as toWikibase is a string for WbDataPage."""
        return hash(self.toWikibase())

    def toWikibase(self) -> str:
        """Convert the data to the value required by the Wikibase API.

        :return: title of the data page incl. namespace
        """
        return self.page.title()

    @classmethod
    def fromWikibase(cls, page_name: str, site: DataSite | None) -> WbDataPage:
        """Create a WbDataPage from the JSON data given by the Wikibase API.

        :param page_name: page name from Wikibase value
        :param site: The Wikibase site
        """
        # TODO: This method signature does not match our parent class (which
        # takes a dictionary argument rather than a string). We should either
        # change this method's signature or rename this method.

        data_site = cls._get_data_site(site)
        page = pywikibot.Page(data_site, page_name)
        return cls(page, site)


class WbGeoShape(WbDataPage):

    """A Wikibase geo-shape representation."""

    @classmethod
    def _get_data_site(cls, site: DataSite) -> APISite:
        """Return the site serving as a geo-shape repository.

        :param site: The Wikibase site
        """
        return site.geo_shape_repository()

    @classmethod
    def _get_type_specifics(cls, site: DataSite) -> dict[str, Any]:
        """Return the specifics for WbGeoShape.

        :param site: The Wikibase site
        """
        return {
            'ending': '.map',
            'label': 'geo-shape',
            'data_site': cls._get_data_site(site),
        }


class WbTabularData(WbDataPage):

    """A Wikibase tabular-data representation."""

    @classmethod
    def _get_data_site(cls, site: DataSite) -> APISite:
        """Return the site serving as a tabular-data repository.

        :param site: The Wikibase site
        """
        return site.tabular_data_repository()

    @classmethod
    def _get_type_specifics(cls, site: DataSite) -> dict[str, Any]:
        """Return the specifics for WbTabularData.

        :param site: The Wikibase site
        """
        return {
            'ending': '.tab',
            'label': 'tabular-data',
            'data_site': cls._get_data_site(site),
        }


class WbUnknown(WbRepresentation):

    """A Wikibase representation for unknown data type.

    This will prevent the bot from breaking completely when a new type
    is introduced.

    This data type is just a json container

    .. versionadded:: 3.0
    .. versionchanged:: 9.4
       *warning* parameter was added
    """

    _items = ('json',)

    def __init__(self, json: dict[str, Any], warning: str = '') -> None:
        """Create a new WbUnknown object.

        :param json: Wikibase JSON
        :param warning: a warning message which is shown once if
            :meth:`toWikibase` is called
        """
        self.json = json
        self.warning = warning

    def toWikibase(self) -> dict[str, Any]:
        """Return the JSON object for the Wikibase API.

        .. versionchanged:: 9.4
           a waning message given by the warning attribute is shown once.

        :return: Wikibase JSON
        """
        if self.warning:
            pywikibot.warning(self.warning)
            self.warning = ''
        return self.json

    @classmethod
    def fromWikibase(cls, data: dict[str, Any],
                     site: DataSite | None = None) -> WbUnknown:
        """Create a WbUnknown from the JSON data given by the Wikibase API.

        :param data: Wikibase JSON
        :param site: The Wikibase site
        """
        return cls(data)



#!/usr/bin/env python3
"""Spawns an interactive Python shell and imports the pywikibot library.

To exit the shell, type :kbd:`ctrl-D` (Linux) or :kbd:`ctrl-Z` (Windows)
or use the :func:`exit` function.

The following local option is supported:

-noimport  Do not import the pywikibot library. All other arguments are
           ignored in this case.

Usage::

    python pwb.py shell [args]

.. versionchanged:: 7.0
   moved to pywikibot.scripts
"""
# (C) Pywikibot team, 2014-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import code
import sys


def main(*args: str) -> None:
    """Script entry point.

    .. versionchanged:: 8.2
       *exitmsg* was added for :func:`code.interact`.
    """
    args = list(args)
    if '-noimport' in args:
        args.remove('-noimport')
        env = None
        warn_type = 'Ignoring'
    else:
        import pywikibot
        args = pywikibot.handle_args(args)
        env = {'pywikibot': pywikibot}
        warn_type = 'Unknown'

    if args:
        print('{} arguments: {}\n'  # noqa: T201
              .format(warn_type, ', '.join(args)))

    # Various stuffs in Python 3.4+, such as history file.
    # This is defined in the site module of the Python Standard Library,
    # and usually called by the built-in CPython interactive shell.
    if hasattr(sys, '__interactivehook__'):
        sys.__interactivehook__()

    code.interact('Welcome to the Pywikibot interactive shell!', local=env,
                  exitmsg='Thank you for using Pywikibot; exiting now...\n')


if __name__ == '__main__':
    if sys.platform == 'win32':
        import platform
        import subprocess
        subprocess.run(f'title Python {platform.python_version()} Shell',
                       shell=True, check=True)
        del subprocess
        del platform
    args = []
    if sys.argv and sys.argv[0].endswith(('shell', 'shell.py')):
        args = sys.argv[1:]
    main(*args)



"""User interfaces module.

The user interface object must define its own `init_handlers()` method
which takes the root logger as its only argument, and which adds to that
logger whatever handlers and formatters are needed to process output and
display it to the user. The default
(:mod:`terminal<userinterfaces.terminal_interface_base>`)
interface sends level :const:`STDOUT` to `sys.stdout` (as all interfaces
should) and sends all other levels to `sys.stderr`; levels
:const:`WARNING` and above are labeled with the level name.

UserInterface objects must also define methods `input()`,
`input_choice()`, `input_list_choice()`, `output()` and `editText()`,
all of which are documented in the abstract class
:class:`userinterfaces._interface_base.ABUIC`.
"""
from __future__ import annotations


#
# (C) Pywikibot team, 2008-2022
#
# Distributed under the terms of the MIT license.
#



"""Collections datatypes."""
#
# (C) Pywikibot team, 2014-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import collections
from abc import ABC, abstractmethod
from collections.abc import Collection, Generator, Iterator, Mapping
from contextlib import suppress
from itertools import chain
from types import TracebackType
from typing import Any, NamedTuple

from pywikibot.backports import Generator as GeneratorType
from pywikibot.exceptions import ArgumentDeprecationWarning
from pywikibot.tools import (
    PYTHON_VERSION,
    deprecated_args,
    issue_deprecation_warning,
)


__all__ = (
    'CombinedError',
    'DequeGenerator',
    'EmptyDefault',
    'GeneratorWrapper',
    'RateLimit',
    'SizedKeyCollection',
    'EMPTY_DEFAULT',
)


class SizedKeyCollection(Collection):

    """Structure to hold values where the key is given by the value itself.

    A structure like a defaultdict but the key is given by the value
    itself and cannot be assigned directly. It returns the number of all
    items with len() but not the number of keys.

    Samples:

        >>> from pywikibot.tools.collections import SizedKeyCollection
        >>> data = SizedKeyCollection('title')
        >>> data.append('foo')
        >>> data.append('bar')
        >>> data.append('Foo')
        >>> list(data)
        ['foo', 'Foo', 'bar']
        >>> len(data)
        3
        >>> 'Foo' in data
        True
        >>> 'foo' in data
        False
        >>> data['Foo']
        ['foo', 'Foo']
        >>> list(data.keys())
        ['Foo', 'Bar']
        >>> data.remove_key('Foo')
        >>> list(data)
        ['bar']
        >>> data.clear()
        >>> list(data)
        []

    .. versionadded:: 6.1
    """

    def __init__(self, keyattr: str) -> None:
        """Initializer.

        :param keyattr: an attribute or method of the values to be hold
            with this collection which will be used as key.
        """
        self.keyattr = keyattr
        self.data: dict[Any, list[Any]] = {}
        self.size = 0

    def __contains__(self, key) -> bool:
        return key in self.data

    def __getattr__(self, key):
        """Delegate Mapping methods to self.data."""
        if key in ('keys', 'values', 'items'):
            return getattr(self.data, key)
        return super().__getattr__(key)

    def __getitem__(self, key) -> list:
        return self.data[key]

    def __iter__(self):
        """Iterate through all items of the tree."""
        yield from chain.from_iterable(self.data.values())

    def __len__(self) -> int:
        """Return the number of all values."""
        return self.size

    def __repr__(self) -> str:
        return str(self.data).replace('defaultdict', self.__class__.__name__)

    def append(self, value) -> None:
        """Add a value to the collection."""
        key = getattr(value, self.keyattr)
        if callable(key):
            key = key()
        if key not in self.data:
            self.data[key] = []
        self.data[key].append(value)
        self.size += 1

    def remove(self, value) -> None:
        """Remove a value from the container."""
        key = getattr(value, self.keyattr)
        if callable(key):
            key = key()
        with suppress(ValueError):
            self.data[key].remove(value)
            self.size -= 1

    def remove_key(self, key) -> None:
        """Remove all values for a given key."""
        with suppress(KeyError):
            self.size -= len(self.data[key])
            del self.data[key]

    def clear(self) -> None:
        """Remove all elements from SizedKeyCollection."""
        self.data.clear()  # defaultdict fails (T282865)
        self.size = 0

    def filter(self, key):
        """Iterate over items for a given key."""
        with suppress(KeyError):
            yield from self.data[key]

    def iter_values_len(self):
        """Yield key, len(values) pairs."""
        for key, values in self.data.items():
            yield key, len(values)


class CombinedError(KeyError, IndexError):

    """An error that gets caught by both KeyError and IndexError.

    .. versionadded:: 3.0
    """


class EmptyDefault(str, Mapping):

    """A default for a not existing siteinfo property.

    It should be chosen if there is no better default known. It acts like an
    empty collections, so it can be iterated through it safely if treated as a
    list, tuple, set or dictionary. It is also basically an empty string.

    Accessing a value via __getitem__ will result in a combined KeyError and
    IndexError.

    .. versionadded:: 3.0
    .. versionchanged:: 6.2
       ``empty_iterator()`` was removed in favour of ``iter()``.
    """

    def __init__(self) -> None:
        """Initialise the default as an empty string."""
        str.__init__(self)

    def __iter__(self):
        """An iterator which does nothing and drops the argument."""
        return iter(())

    def __getitem__(self, key):
        """Raise always a :py:obj:`CombinedError`."""
        raise CombinedError(key)


#:
EMPTY_DEFAULT = EmptyDefault()


class DequeGenerator(Iterator, collections.deque):

    """A generator that allows items to be added during generating.

    .. versionadded:: 3.0
    .. versionchanged:: 6.1
       Provide a representation string.
    """

    def __next__(self):
        """Iterator method."""
        if self:
            return self.popleft()
        raise StopIteration

    def __repr__(self) -> str:
        """Provide an object representation without clearing the content."""
        items = list(self)
        result = f'{self.__class__.__name__}({items})'
        self.extend(items)
        return result


class GeneratorWrapper(ABC, Generator):

    """A Generator base class which wraps the internal `generator` property.

    This generator iterator also has :python:`generator.close()
    <reference/expressions.html#generator.close>` mixin method and it can
    be used as Iterable and Iterator as well.

    .. versionadded:: 7.6

    Example:

    >>> class Gen(GeneratorWrapper):
    ...     @property
    ...     def generator(self):
    ...         return (c for c in 'Pywikibot')
    >>> gen = Gen()
    >>> next(gen)  # can be used as Iterator ...
    'P'
    >>> next(gen)
    'y'
    >>> ''.join(c for c in gen)  # ... or as Iterable
    'wikibot'
    >>> next(gen)  # the generator is exhausted ...
    Traceback (most recent call last):
        ...
    StopIteration
    >>> gen.restart()  # ... but can be restarted
    >>> next(gen) + next(gen)
    'Py'
    >>> gen.close()  # the generator may be closed
    >>> next(gen)
    Traceback (most recent call last):
        ...
    StopIteration
    >>> gen.restart()  # restart a closed generator
    >>> # also send() and throw() works
    >>> gen.send(None) + gen.send(None)
    'Py'
    >>> gen.throw(RuntimeError('Foo'))
    Traceback (most recent call last):
        ...
    RuntimeError: Foo

    .. seealso:: :pep:`342`
    """

    @property
    @abstractmethod
    def generator(self) -> GeneratorType[Any, Any, Any]:
        """Abstract generator property."""
        yield from ()

    def send(self, value: Any) -> Any:
        """Return next yielded value from generator or raise StopIteration.

        The `value` parameter is ignored yet; usually it should be ``None``.
        If the wrapped generator property exits without yielding another
        value this method raises `StopIteration`. The send method works
        like the `next` function with a GeneratorWrapper instance as
        parameter.

        Refer :python:`generator.send()
        <reference/expressions.html#generator.send>` for its usage.

        :raises TypeError: generator property is not a generator
        """
        if not isinstance(self.generator, GeneratorType):
            raise TypeError('generator property is not a generator but '
                            f'{type(self.generator).__name__}')
        if not hasattr(self, '_started_gen'):
            # start the generator
            self._started_gen = self.generator
        return next(self._started_gen)

    @deprecated_args(val='value', tb='traceback')  # since 10.7.0
    def throw(self,
              typ: BaseException | type[BaseException] | None = None,
              value: Any = None,
              traceback: TracebackType | None = None) -> None:
        """Raise an exception inside the wrapped generator.

        Refer :python:`generator.throw()
        <reference/expressions.html#generator.throw>` for various
        parameter usage.

        .. versionchanged:: 10.7
           The *val* and *tb* parameters were renamed to *value* and
           *traceback*.
        .. deprecated:: 10.7
           The ``(type, value, traceback)`` signature is deprecated; use
           single-arg signature ``throw(value)`` instead.

        :raises RuntimeError: No generator started
        :raises TypeError: Invalid type for *typ* argument
        """
        if not hasattr(self, '_started_gen'):
            raise RuntimeError('No generator was started')

        # New-style (single exception instance) with keyword argument
        if typ is None and traceback is None and isinstance(value,
                                                            BaseException):
            self._started_gen.throw(value)
            return

        if PYTHON_VERSION > (3, 8) and not (value is None
                                            and traceback is None):
            # Old-style (type, value, traceback) signature
            issue_deprecation_warning(
                'The (type, value, traceback) signature of throw()',
                'the single-arg signature',
                warning_class=ArgumentDeprecationWarning,
                since='10.7.0'
            )
            self._started_gen.throw(typ, value, traceback)
            return

        # New-style (single exception instance)
        self._started_gen.throw(typ)

    def restart(self) -> None:
        """Restart the generator."""
        with suppress(AttributeError):
            del self._started_gen


class RateLimit(NamedTuple):

    """A namedtuple which can hold rate limit content.

    This class is used by :meth:`APISite.ratelimit()
    <pywikibot.site._apisite.APISite.ratelimit>`.

    .. note:: :meth:`delay` and :meth:`ratio` properties cannot be
       sliced or used with tuple indices. They must be used as
       attributes.

    >>> limit = RateLimit('user', 500, 10)
    >>> limit
    RateLimit(group='user', hits=500, seconds=10)
    >>> limit.delay
    0.02
    >>> limit.ratio
    50.0
    >>> limit._fields
    ('group', 'hits', 'seconds')
    >>> limit._asdict()  # doctest: +SKIP
    {'group': 'user', 'hits': 500, 'seconds': 10}
    >>> limit[0]
    'user'
    >>> limit[-1]
    10
    >>> user, hits, seconds = limit
    >>> hits, seconds
    (500, 10)
    >>> newlimit = limit._replace(seconds=0)
    >>> newlimit.delay
    0.0
    >>> newlimit.ratio
    inf

    .. versionadded:: 9.0
    """

    group: str = 'unknown'
    hits: int = 50
    seconds: int = 0

    @property
    def delay(self) -> float:
        """Calculate a delay value which is the inverse of :meth:`ratio`."""
        return self.seconds / self.hits

    @property
    def ratio(self) -> float:
        """Calculate a ratio how many hits can be done within one second."""
        return self.hits / self.seconds if self.seconds != 0 else float('inf')



"""Decorators for Page objects."""
#
# (C) Pywikibot team, 2017-2024
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import pywikibot
from pywikibot.exceptions import (
    Error,
    OtherPageSaveError,
    PageSaveRelatedError,
)
from pywikibot.tools import SPHINX_RUNNING, add_full_name, manage_wrapping


def allow_asynchronous(func):
    """Decorator to make it possible to run a BasePage method asynchronously.

    This is done when the method is called with kwarg
    :code:`asynchronous=True`. Optionally, you can also provide kwarg
    callback, which, if provided, is a callable that gets the page as
    the first and a possible exception that occurred during saving in
    the second thread or None as the second argument.
    """
    def handle(func, self, *args, **kwargs) -> None:
        do_async = kwargs.pop('asynchronous', False)
        callback = kwargs.pop('callback', None)
        err = None
        try:
            func(self, *args, **kwargs)
        # TODO: other "expected" error types to catch?
        except Error as edit_err:
            err = edit_err  # edit_err will be deleted in the end of the scope
            link = self.title(as_link=True)
            if do_async:
                pywikibot.error(f'page {link} not saved due to {err}\n')
            pywikibot.log(f'Error saving page {link} ({err})\n',
                          exc_info=True)
            if not callback and not do_async:
                if isinstance(err, PageSaveRelatedError):
                    raise err
                raise OtherPageSaveError(self, err)
        if callback:
            callback(self, err)

    def wrapper(self, *args, **kwargs) -> None:
        if kwargs.get('asynchronous'):
            pywikibot.async_request(handle, func, self, *args, **kwargs)
        else:
            handle(func, self, *args, **kwargs)

    manage_wrapping(wrapper, func)

    return wrapper


if not SPHINX_RUNNING:
    # T365286: decorate allow_asynchronous function with add_full_name
    allow_asynchronous = add_full_name(allow_asynchronous)



"""Family module for Wikidata."""
#
# (C) Pywikibot team, 2012-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from pywikibot import config, family


class Family(family.WikimediaFamily, family.DefaultWikibaseFamily):

    """Family class for Wikidata."""

    name = 'wikidata'

    langs = {
        'wikidata': 'www.wikidata.org',
        'test': 'test.wikidata.org',
        'beta': 'wikidata.beta.wmcloud.org',
    }

    # Sites we want to edit but not count as real languages
    test_codes = ['test', 'beta']

    interwiki_forward = 'wikipedia'

    category_redirect_templates = {
        'wikidata': (
            'Category redirect',
        ),
    }

    # Subpages for documentation.
    doc_subpages = {
        '_default': (('/doc', ), ['wikidata']),
    }

    # Disable cosmetic changes
    config.cosmetic_changes_disable.update({
        'wikidata': ('wikidata', 'test', 'beta')
    })



"""Server-Sent Events client.

This file is part of the Pywikibot framework.

This module requires requests-sse to be installed::

    pip install "requests-sse>=0.5.0"

.. versionadded:: 3.0
.. versionchanged:: 10.0
   ``requests-sse`` package is required instead of ``sseclient``.
"""
#
# (C) Pywikibot team, 2017-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import json
from datetime import timedelta
from functools import partial
from typing import Any

from requests.packages.urllib3.exceptions import ProtocolError
from requests.packages.urllib3.util.response import httplib

from pywikibot import Site, Timestamp, config, debug, warning
from pywikibot.backports import Dict, List, NoneType
from pywikibot.comms.http import user_agent
from pywikibot.tools import cached, deprecated_args
from pywikibot.tools.collections import GeneratorWrapper


try:
    from requests_sse import EventSource
except ModuleNotFoundError as e:
    EventSource = e


INSTALL_MSG = """requests-sse is required for EventStreams;
install it with

    pip install "requests-sse>=0.5.0"
"""


class EventStreams(GeneratorWrapper):

    """Generator class for Server-Sent Events (SSE) protocol.

    It provides access to arbitrary streams of data including recent changes.

    Usage:

    >>> stream = EventStreams(streams='recentchange')
    >>> stream.register_filter(type='edit', wiki='wikidatawiki', bot=True)
    >>> change = next(stream)
    >>> msg = '{type} on page {title} by {user}.'.format_map(change)
    >>> print(msg)  # doctest: +SKIP
    edit on page Q2190037 by KrBot.
    >>> from pprint import pprint
    >>> pprint(change, width=75)  # doctest: +SKIP
    {'$schema': '/mediawiki/recentchange/1.0.0',
     'bot': True,
     'comment': '/* wbsetreference-set:2| */ [[Property:P10585]]: 96FPN, см. '
                '/ see [[Template:Autofix|autofix]] на / on [[Property '
                'talk:P356]]',
     'id': 1728475074,
     'length': {'new': 8871, 'old': 8871},
     'meta': {'domain': 'www.wikidata.org',
              'dt': '2022-07-12T17:54:15Z',
              'id': '2cdec62f-a2b3-49b8-9a52-85a42236fb99',
              'offset': 4000957901,
              'partition': 0,
              'request_id': 'f7896e77-fd2b-4a95-a9e4-44c1e3ad818b',
              'stream': 'mediawiki.recentchange',
              'topic': 'eqiad.mediawiki.recentchange',
              'uri': 'https://www.wikidata.org/wiki/Q2190037'},
     'minor': False,
     'namespace': 0,
     'parsedcomment': '\u200e<span dir="auto"><span '
                      'class="autocomment">Изменена ссылка на заявление: '
                      '</span></span> <a href="/wiki/Property:P10585" '
                      'title="Property:P10585">Property:P10585</a>: 96FPN, '
                      'см. / see <a href="/wiki/Template:Autofix" '
                      'title="Template:Autofix">autofix</a> на / on <a '
                      'href="/wiki/Property_talk:P356" title="Property '
                      'talk:P356">Property talk:P356</a>',
     'patrolled': True,
     'revision': {'new': 1676015019, 'old': 1675697125},
     'server_name': 'www.wikidata.org',
     'server_script_path': '/w',
     'server_url': 'https://www.wikidata.org',
     'timestamp': 1657648455,
     'title': 'Q2190037',
     'type': 'edit',
     'user': 'KrBot',
     'wiki': 'wikidatawiki'}
    >>> pprint(next(stream), width=75)  # doctest: +ELLIPSIS
    {'$schema': '/mediawiki/recentchange/1.0.0',
     'bot': True,
     ...
     'server_name': 'www.wikidata.org',
     'server_script_path': '/w',
     'server_url': 'https://www.wikidata.org',
     ...
     'type': 'edit',
     'user': ...,
     'wiki': 'wikidatawiki'}
    >>> del stream

    .. versionchanged:: 7.6
       subclassed from :class:`tools.collections.GeneratorWrapper`.
    .. versionchanged:: 10.0
       *retry* value is doubled for each consecutive connect try.
    """

    @deprecated_args(last_id='last_event_id')  # since 10.0.0
    def __init__(self, **kwargs) -> None:
        """Initializer.

        :keyword bool canary: if True, include canary events, see
            https://w.wiki/7$2z for more info.
        :keyword APISite site: a project site object. Used if no *url*
            is given.
        :keyword int retry: Number of milliseconds to wait after disconnects
            before attempting to reconnect. The server may change this
            by including a 'retry' line in a message. Retries are handled
            automatically.

            .. versionchanged:: 10.0
               5 seconds are used instead of 3 seconds as default.

        :keyword pywikibot.Timestamp | str since: a timestamp for older
            events; there will likely be between 7 and 31 days of
            history available but is not guaranteed. It may be given as
            a pywikibot.Timestamp, an ISO 8601 string or a mediawiki
            timestamp string.
        :keyword Iterable[str] | str streams: event stream types.
            Mandatory when no url is given. Multiple streams may be
            given as a string with comma separated stream types or an
            iterable of strings
        :keyword int | float | tuple[int | float, int | float] timeout:
            a timeout value indication how long to wait to send data
            before giving up
        :keyword str url: an url retrieving events from. Will be set up
            to a default url using _site.family settings, stream types
            and timestamp

        :keyword Any last_event_id: [*requests-sse*] If provided, this
            parameter will be sent to the server to tell it to return
            only messages more recent than this ID.
        :keyword requests.Session session: [*requests-sse*] specifies a
            requests.Session, if not, create a default requests.Session.
        :keyword Callable[[], None] on_open: [*requests-sse*] event
            handler for open event
        :keyword Callable[[requests_sse.MessageEvent], None] on_message:
            [*requests-sse*] event handler for message event
        :keyword Callable[[], None] on_error: [*requests-sse*] event
            handler for error event
        :keyword int chunk_size: [*requests*] A maximum size of the chunk
            for chunk-encoded requests.

            .. versionchanged:: 10.0
               None is used instead of 1024 as default value.

        :param kwargs: Other keyword arguments passed to `requests_sse`
            and `requests` library
        :raises ModuleNotFoundError: requests-sse is not installed
        :raises NotImplementedError: no stream types specified

        .. seealso:: https://stream.wikimedia.org/?doc#streams for
           available Wikimedia stream types to be passed with `streams`
           parameter.
        .. note:: *retry* keyword argument is used instead of the
           underlying *reconnection_time* argument which is ignored.
        """
        if isinstance(EventSource, ModuleNotFoundError):
            raise ImportError(INSTALL_MSG) from EventSource

        self.filter: Dict[str, List[Any]] = {'all': [], 'any': [], 'none': []}
        self._total: int | None = None
        self._canary = kwargs.pop('canary', False)

        try:
            self._site = kwargs.pop('site')
        except KeyError:  # T335720
            self._site = Site()

        self._streams = kwargs.pop('streams', None)
        if self._streams and not isinstance(self._streams, str):
            self._streams = ','.join(self._streams)

        self._since = kwargs.pop('since', None)
        if self._since:
            # assume this is a mw timestamp, convert it to a Timestamp object
            if isinstance(self._since, str) and '-' not in self._since:
                self._since = Timestamp.fromtimestampformat(self._since)
            if isinstance(self._since, Timestamp):
                self._since = self._since.isoformat()

        self._url = kwargs.get('url') or self.url
        kwargs.setdefault('url', self._url)

        retry = kwargs.pop('retry', None)
        if retry:
            kwargs['reconnection_time'] = timedelta(milliseconds=retry)

        kwargs.setdefault('timeout', config.socket_timeout)

        kwargs.setdefault('headers', {})
        kwargs['headers'].setdefault('user-agent', user_agent(self._site))

        self.sse_kwargs = kwargs

    def __repr__(self) -> str:
        """Return representation string."""
        kwargs = self.sse_kwargs.copy()
        if self._site != Site():
            kwargs['site'] = self._site
        if self._streams:
            kwargs['streams'] = self._streams
            kwargs.pop('url')
        if self._since:
            kwargs['since'] = self._since
        if kwargs['timeout'] == config.socket_timeout:
            kwargs.pop('timeout')
        return '{}({})'.format(type(self).__name__, ', '.join(
            f'{k}={v!r}' for k, v in kwargs.items()))

    @property
    @cached
    def url(self) -> str:
        """Get the EventStream's url.

        :raises NotImplementedError: no stream types specified
        """
        if self._streams is None:
            raise NotImplementedError(
                f'No streams specified for class {type(self).__name__}')
        return '{host}{path}/{streams}{since}'.format(
            host=self._site.eventstreams_host(),
            path=self._site.eventstreams_path(),
            streams=self._streams,
            since=f'?since={self._since}' if self._since else '')

    def set_maximum_items(self, value: int | None) -> None:
        """Set the maximum number of items to be retrieved from the stream.

        If not called, most queries will continue as long as there is
        more data to be retrieved from the stream.

        :param value: The value of maximum number of items to be
            retrieved in total to set.
        """
        if value is not None:
            self._total = int(value)
            debug(f'{type(self).__name__}: Set limit (maximum_items) to '
                  f'{self._total}.')

    def register_filter(self, *args, **kwargs) -> None:
        """Register a filter.

        Filter types:

        There are 3 types of filter: 'all', 'any' and 'none'.
        The filter type must be given with the keyword argument 'ftype'
        (see below). If no 'ftype' keyword argument is given, 'all' is
        assumed as default.

        You may register multiple filters for each type of filter.
        The behaviour of filter type is as follows:

        - **'none'**: Skip if the any filter matches. Otherwise check 'all'.
        - **'all'**: Skip if not all filter matches. Otherwise check 'any':
        - **'any'**: Skip if no given filter matches. Otherwise pass.

        Filter functions:

        Filter may be specified as external function methods given as
        positional argument like::

            def foo(data):
                return True

            register_filter(foo, ftype='any')

        The data dict from event is passed to the external filter function as
        a parameter and that method must handle it in a proper way and return
        ``True`` if the filter matches and ``False`` otherwise.

        Filter keys and values:

        Another method to register a filter is to pass pairs of keys and values
        as keyword arguments to this method. The key must be a key of the event
        data dict and the value must be any value or an iterable of values the
        ``data['key']`` may match or be part of it. Samples::

            register_filter(server_name='de.wikipedia.org')  # 1
            register_filter(type=('edit', 'log'))  # 2
            register_filter(ftype='none', bot=True)  # 3

        Explanation for the result of the filter function:

        1. ``return data['sever_name'] == 'de.wikipedia.org'``
        2. ``return data['type'] in ('edit', 'log')``
        3. ``return data['bot'] is True``

        :keyword ftype: The filter type, one of 'all', 'any', 'none'.
            Default value is 'all'
        :type ftype: str
        :param args: You may pass your own filter functions here.
            Every function should be able to handle the data dict from events.
        :type args: callable
        :param kwargs: Any key returned by event data with an event data value
            for this given key.
        :type kwargs: str, list, tuple or other sequence
        :raise TypeError: A given args parameter is not a callable.
        """
        def _is(data, key=None, value=None):
            return key in data and data[key] is value

        def _eq(data, key=None, value=None):
            return key in data and data[key] == value

        def _in(data, key=None, value=None):
            return key in data and data[key] in value

        ftype = kwargs.pop('ftype', 'all')  # set default ftype value

        # register an external filter function
        for func in args:
            if not callable(func):
                raise TypeError(f'{func} is not a callable')

            self.filter[ftype].append(func)

        # register pairs of keys and items as a filter function
        for key, value in kwargs.items():
            # append function for singletons
            if isinstance(value, (bool, NoneType)):
                self.filter[ftype].append(partial(_is, key=key, value=value))
            # append function for a single value
            elif isinstance(value, (str, int)):
                self.filter[ftype].append(partial(_eq, key=key, value=value))
            # append function for an iterable as value
            else:
                self.filter[ftype].append(partial(_in, key=key, value=value))

    def streamfilter(self, data: dict[str, Any]) -> bool:
        """Filter function for eventstreams.

        See the description of register_filter() how it works.

        :param data: event data dict used by filter functions
        """
        if not self._canary and data.get('meta', {}).get('domain') == 'canary':
            return False  # T266798

        if any(function(data) for function in self.filter['none']):
            return False

        if not all(function(data) for function in self.filter['all']):
            return False

        if not self.filter['any']:
            return True

        return any(function(data) for function in self.filter['any'])

    @property
    def generator(self):
        """Inner generator.

        .. versionchanged:: 7.6
           changed from iterator method to generator property
        """
        n = 0
        event = None
        while self._total is None or n < self._total:
            if not hasattr(self, 'source'):
                self.source = EventSource(**self.sse_kwargs)
                self.source.connect(config.max_retries)

            try:
                event = next(self.source)
            except (ProtocolError, OSError, httplib.IncompleteRead) as e:
                warning(
                    f'Connection error: {e}.\nTry to re-establish connection.')
                self.source.close()
                del self.source
                if event is not None:
                    self.sse_kwargs['latest_event_id'] = event.last_event_id
                continue

            if event.type == 'message':
                if event.data:
                    try:
                        element = json.loads(event.data)
                    except ValueError as e:
                        warning(f'Could not load json data from\n{event}\n{e}')
                    else:
                        if self.streamfilter(element):
                            n += 1
                            yield element
                # else: ignore empty message
            elif event.type == 'error':
                warning(f'Encountered error: {event.data}')
            else:
                warning(f'Unknown event {event.type} occurred.')

        debug(f'{type(self).__name__}: Stopped iterating due to exceeding item'
              ' limit.')

        self.source.close()
        del self.source


def site_rc_listener(site, total: int | None = None):
    """Yield changes received from EventStream.

    :param site: the Pywikibot.Site object to yield live recent changes
        for
    :type site: Pywikibot.BaseSite
    :param total: the maximum number of changes to return
    :return: pywikibot.comms.eventstream.rc_listener configured for
        given site
    :raises ModuleNotFoundError: requests-sse installation is required
    """
    if isinstance(EventSource, ModuleNotFoundError):
        raise ModuleNotFoundError(INSTALL_MSG) from EventSource

    stream = EventStreams(streams='recentchange', site=site)
    stream.set_maximum_items(total)
    stream.register_filter(server_name=site.hostname())
    return stream



"""Platform independent terminal interface module.

It imports the appropriate operating system specific implementation.
"""
#
# (C) Pywikibot team, 2003-2023
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import sys


if sys.platform == 'win32':
    from pywikibot.userinterfaces.terminal_interface_win32 import Win32UI as UI
else:
    from pywikibot.userinterfaces.terminal_interface_unix import UnixUI as UI

__all__ = ('UI',)



"""**Pywikibot Exceptions** and warning classes.

This module contains all exception and warning classes used throughout
the framework::

    Exception
     +-- Error
          +-- APIError
          |    +-- APIMWError
          |    +-- UploadError
          +-- AutoblockUserError
          +-- CaptchaError
          +-- ClientError
          |    +-- Client414Error
          +-- InvalidTitleError
          +-- NoUsernameError
          +-- PageInUseError
          +-- PageRelatedError
          |    +-- CircularRedirectError
          |    +-- InterwikiRedirectPageError
          |    +-- IsNotRedirectPageError
          |    +-- IsRedirectPageError
          |    +-- NoMoveTargetError
          |    +-- NoPageError
          |    +-- NoRenameTargetError
          |    +-- NotEmailableError
          |    +-- PageLoadRelatedError
          |    |    +-- InconsistentTitleError
          |    |    +-- InvalidPageError
          |    |    +-- NoSiteLinkError
          |    +-- PageSaveRelatedError
          |    |    +-- EditConflictError
          |    |    |    +-- ArticleExistsConflictError
          |    |    |    +-- PageCreatedConflictError
          |    |    |    +-- PageDeletedConflictError
          |    |    +-- LockedPageError
          |    |    |    +-- LockedNoPageError
          |    |    |    +-- CascadeLockedPageError
          |    |    +-- NoCreateError
          |    |    +-- OtherPageSaveError
          |    |    +-- SpamblacklistError
          |    |    +-- TitleblacklistError
          |    |    +-- AbuseFilterDisallowedError
          |    +-- UnsupportedPageError
          +-- SectionError
          +-- ServerError
          |    +-- FatalServerError
          |    +-- Server504Error
          +-- SiteDefinitionError
          |    +-- UnknownFamilyError
          |    +-- UnknownSiteError
          +-- TimeoutError
          |    +-- MaxlagTimeoutError
          +-- TranslationError
          +-- UserRightsError
          |    +-- HiddenKeyError (KeyError)
          +-- UnknownExtensionError (NotImplementedError)
          +-- VersionParseError
          +-- WikiBaseError
               +-- CoordinateGlobeUnknownError (NotImplementedError)
               +-- EntityTypeUnknownError
               +-- NoWikibaseEntityError

    UserWarning
     +-- ArgumentDeprecationWarning (FutureWarning)
     +-- FamilyMaintenanceWarning

    RuntimeWarning
     +-- NotImplementedWarning


Error: Base class, all exceptions should the subclass of this class.

  - CaptchaError: Captcha is asked and config.solve_captcha == False
  - ClientError: A problem with the client request
  - AutoblockUserError: requested action on a virtual autoblock user not valid
  - InvalidTitleError: Invalid page title
  - NoUsernameError: Username is not in user config file, or it is invalid.
  - PageInUseError: Page cannot be reserved due to a lock
  - SectionError: The section specified by # does not exist
  - TranslationError: no language translation found, i18n/l10n message
    not available
  - UnknownExtensionError: Extension is not defined for this site
  - UserRightsError: insufficient rights for requested action
  - VersionParseError: failed to parse version information

APIError: wiki API returned an error

  - APIMWError: MediaWiki internal exception
  - UploadError: upload failed

SiteDefinitionError: Site loading problem

  - UnknownSiteError: Site does not exist in Family
  - UnknownFamilyError: Family is not registered

PageRelatedError: any exception which is caused by an operation on a Page.

  - NoPageError: Page does not exist
  - UnsupportedPageError: Page is not supported due to a namespace restriction
  - IsRedirectPageError: Page is a redirect page
  - IsNotRedirectPageError: Page is not a redirect page
  - CircularRedirectError: Page is a circular redirect
  - InterwikiRedirectPageError: Page is a redirect to another site
  - InvalidPageError: Page is invalid e.g. without history
  - NotEmailableError: The target user has disabled email
  - NoMoveTargetError: An expected move target page does not exist
  - NoRenameTargetError: An expected rename target user does not exist

PageLoadRelatedError: any exception which happens while loading a Page.
  - InconsistentTitleError: Page receives a title inconsistent with query
  - NoSiteLinkError: ItemPage has no sitelink to given site

PageSaveRelatedError: page exceptions within the save operation on a Page

  - AbuseFilterDisallowedError: AbuseFilter disallowed
  - SpamblacklistError: MediaWiki spam filter detected a blacklisted URL
  - TitleblacklistError: MediaWiki detected a blacklisted page title
  - OtherPageSaveError: misc. other save related exception.
  - LockedPageError: Page is locked
      - LockedNoPageError: Title is locked against creation
      - CascadeLockedPageError: Page is locked due to cascading protection
  - EditConflictError: Edit conflict while uploading the page
      - PageDeletedConflictError: Page was deleted since being retrieved
      - PageCreatedConflictError: Page was created by another user
      - ArticleExistsConflictError: Page article already exists
  - NoCreateError: parameter nocreate not allow page creation

ServerError: a problem with the server.

  - FatalServerError: A fatal/non-recoverable server error
  - Server414Error: Server timed out with HTTP 414 code
  - Server504Error: Server timed out with HTTP 504 code

WikiBaseError: any issue specific to Wikibase.

  - NoWikibaseEntityError: entity doesn't exist
  - CoordinateGlobeUnknownError: globe is not implemented yet.
  - EntityTypeUnknownError: entity type is not available on the site.

TimeoutError: request failed with a timeout

  - MaxlagTimeoutError: request failed with a maxlag timeout

DeprecationWarning: old functionality replaced by new functionality

PendingDeprecationWarning: problematic code which has not yet been
fully deprecated, possibly because a replacement is not available

RuntimeWarning: problems developers should have fixed, and users need to
be aware of its status.

  - NotImplementedWarning: functionality not implemented

UserWarning: warnings targeted at users

  - config._ConfigurationDeprecationWarning: user configuration file problems
  - login._PasswordFileWarning: password file problems
  - ArgumentDeprecationWarning: command line argument problems
  - FamilyMaintenanceWarning: missing information in family definition

.. versionchanged:: 6.0
   exceptions were renamed and are ending with "Error".

.. versionchanged:: 7.0
   All Pywikibot Error exceptions must be imported from
   ``pywikibot.exceptions``. Deprecated exceptions identifiers were
   removed.

.. versionchanged:: 8.1
   ``Server414Error`` class is deprecated; use :class:`Client414Error`
   instead.
"""
#
# (C) Pywikibot team, 2008-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import re
from typing import Any

import pywikibot
from pywikibot.tools import ModuleDeprecationWrapper
from pywikibot.tools._deprecate import _NotImplementedWarning


class NotImplementedWarning(_NotImplementedWarning):

    """Feature that is no longer implemented."""


class ArgumentDeprecationWarning(UserWarning, FutureWarning):

    """Command line argument that is no longer supported."""


class FamilyMaintenanceWarning(UserWarning):

    """Family class is missing definitions."""


class Error(Exception):

    """Pywikibot error."""

    def __init__(self, arg: Exception | str) -> None:
        """Initializer."""
        self.unicode = str(arg)

    def __str__(self) -> str:
        """Return a string representation."""
        return self.unicode


class APIError(Error):

    """The wiki site returned an error message."""

    def __init__(self, code: str, info: str, **kwargs: Any) -> None:
        """Save error dict returned by MW API."""
        self.code = code
        self.info = info
        self.other = kwargs
        self.unicode = str(self)

    def __repr__(self) -> str:
        """Return internal representation."""
        return '{name}("{code}", "{info}", {other})'.format(
            name=self.__class__.__name__, **self.__dict__)

    def __str__(self) -> str:
        """Return a string representation."""
        if self.other:
            return '{}: {}\n[{}]'.format(
                self.code,
                self.info,
                ';\n '.join(
                    f'{key}: {val}'
                    for key, val in self.other.items()))

        return f'{self.code}: {self.info}'


class APIMWError(APIError):

    """The API site returned an error about a MediaWiki internal exception."""

    def __init__(self, mediawiki_exception_class_name: str, info: str,
                 **kwargs: Any) -> None:
        """Save error dict returned by MW API."""
        self.mediawiki_exception_class_name = mediawiki_exception_class_name
        code = 'internal_api_error_' + mediawiki_exception_class_name
        super().__init__(code, info, **kwargs)


class UploadError(APIError):

    """Upload failed with a warning message (passed as the argument)."""

    def __init__(self, code: str, message: str,
                 file_key: str | None = None,
                 offset: int | bool = 0) -> None:
        """Create a new UploadError instance.

        :param file_key: The file_key of the uploaded file to reuse it
            later. If no key is known or it is an incomplete file it may
            be None.
        :param offset: The starting offset for a chunked upload. Is
            False when there is no offset.
        """
        super().__init__(code, message)
        self.file_key = file_key
        self.offset = offset

    @property
    def message(self) -> str:
        """Return warning message."""
        return self.info


class PageRelatedError(Error):

    """Abstract Exception, used when the exception concerns a particular Page.

    This class should be used when the Exception concerns a particular
    Page, and when a generic message can be written once for all.

    .. versionchanged:: 10.5
       A pageid is accepted with the first parameter
    """

    # Preformatted message where the page title will be inserted.
    # Override this in subclasses.
    message = ''

    def __init__(self, page: pywikibot.page.BasePage | int,
                 message: str | None = None) -> None:
        """Initializer.

        :param page: Page object or pageid that caused the exception
        """
        if message:
            self.message = message

        if self.message is None:
            raise Error("PageRelatedError is abstract. Can't instantiate it!")

        self.page = page
        if isinstance(page, pywikibot.page.BasePage):
            self.title = str(page)
            self.site = page.site
        else:
            self.title = f'{page} (pageid)'
            self.site = ''

        if re.search(r'\{\w+\}', self.message):
            msg = self.message.format_map(self.__dict__)
        else:
            msg = self.message.format(self.title)

        super().__init__(msg)


class PageSaveRelatedError(PageRelatedError):

    """Saving the page has failed."""

    message = 'Page {} was not saved.'


class OtherPageSaveError(PageSaveRelatedError):

    """Saving the page has failed due to uncatchable error."""

    message = 'Edit to page {title} failed:\n{reason}'

    def __init__(self, page: pywikibot.page.BasePage,
                 reason: str | Exception) -> None:
        """Initializer.

        :param reason: Details of the problem
        """
        self.reason = reason
        super().__init__(page)

    @property
    def args(self) -> str:  # type: ignore[override]
        """Expose args."""
        return str(self.reason)


class NoUsernameError(Error):

    """Username is not in user config file (user-config.py)."""


class NoPageError(PageRelatedError):

    """Page does not exist."""

    message = "Page {} doesn't exist."


class UnsupportedPageError(PageRelatedError):

    """Unsupported page due to namespace restriction."""

    # namespaces < 0 aren't supported (T169213)
    message = 'Page {} is not supported due to namespace restriction.'


class NoMoveTargetError(PageRelatedError):

    """Expected move target page not found."""

    message = 'Move target page of {} not found.'


class NoRenameTargetError(PageRelatedError):

    """Expected rename target user not found.

    .. versionadded:: 9.4
    """

    message = 'Rename target user of {} not found.'


class PageLoadRelatedError(PageRelatedError):

    """Loading the contents of a Page object has failed."""

    message = 'Page {} was not loaded.'


class InconsistentTitleError(PageLoadRelatedError):

    """Page receives a title inconsistent with query."""

    def __init__(self, page: pywikibot.page.BasePage, actual: str) -> None:
        """Initializer.

        :param page: Page that caused the exception
        :param actual: title obtained by query
        """
        self.message = f"Query on {{}} returned data on '{actual}'"
        super().__init__(page)


class NoSiteLinkError(PageLoadRelatedError, NoPageError):

    """ItemPage has no sitelink to the given site.

    .. versionadded:: 8.1
    .. deprecated:: 8.1
       :exc:`NoPageError` dependency.
    """

    def __init__(self, page: pywikibot.page.ItemPage, dbname: str) -> None:
        """Initializer.

        :param page: ItemPage that caused the exception
        :param dbname: site identifier of the queried sitelink
        """
        self.message = f'Item {{}} has no sitelink to {dbname!r}'
        super().__init__(page)


class SiteDefinitionError(Error):

    """Site does not exist."""


class UnknownSiteError(SiteDefinitionError):

    """Site does not exist in Family."""


class UnknownFamilyError(SiteDefinitionError):

    """Family is not registered."""


class UnknownExtensionError(Error, NotImplementedError):

    """Extension is not defined."""


class VersionParseError(Error):

    """Failed to parse version information."""


class IsRedirectPageError(PageRelatedError):

    """Page is a redirect page."""

    message = 'Page {} is a redirect page.'


class IsNotRedirectPageError(PageRelatedError):

    """Page is not a redirect page."""

    message = 'Page {} is not a redirect page.'


class CircularRedirectError(PageRelatedError):

    """Page is a circular redirect.

    Exception argument is the redirect target; this may be the same
    title as this page or a different title (in which case the target
    page directly or indirectly redirects back to this one)
    """

    message = 'Page {} is a circular redirect.'


class InterwikiRedirectPageError(PageRelatedError):

    """Page is a redirect to another site.

    This is considered invalid in Pywikibot. See bug :phab:`T75184`.
    """

    message = ('Page redirects to a page on another Site.\n'
               'Page: {page}\n'
               'Target page: {target_page} on {target_site}.')

    def __init__(self, page: pywikibot.page.BasePage,
                 target_page: pywikibot.page.BasePage) -> None:
        """Initializer.

        :param target_page: Target page of the redirect.
        """
        self.target_page = target_page
        self.target_site = target_page.site
        super().__init__(page)


class InvalidPageError(PageLoadRelatedError):

    """Missing page history.

    .. versionadded:: 6.2
    """

    message = 'Page {} is invalid.'


class InvalidTitleError(Error):

    """Invalid page title."""


class LockedPageError(PageSaveRelatedError):

    """Page is locked."""

    message = 'Page {} is locked.'


class LockedNoPageError(LockedPageError):

    """Title is locked against creation."""

    message = 'Page {} does not exist and is locked preventing creation.'


class CascadeLockedPageError(LockedPageError):

    """Page is locked due to cascading protection."""

    message = 'Page {} is locked due to cascading protection.'


class SectionError(Error):

    """The section specified by # does not exist."""


class NoCreateError(PageSaveRelatedError):

    """Parameter nocreate doesn't allow page creation."""

    message = 'Page {} could not be created due to parameter nocreate'


class EditConflictError(PageSaveRelatedError):

    """There has been an edit conflict while uploading the page."""

    message = 'Page {} could not be saved due to an edit conflict'


class PageDeletedConflictError(EditConflictError):

    """Page was deleted since being retrieved."""

    message = 'Page {} has been deleted since last retrieved.'


class PageCreatedConflictError(EditConflictError):

    """Page was created by another user."""

    message = 'Page {} has been created since last retrieved.'


class ArticleExistsConflictError(EditConflictError):

    """Page already exists."""

    message = ('Destination article {} already exists and is not a redirect '
               'to the source article')


class AbuseFilterDisallowedError(PageSaveRelatedError):

    """Page save failed because the AbuseFilter disallowed it."""

    message = ('Edit to page {title} disallowed by the AbuseFilter.\n'
               '{info}')

    def __init__(self, page: pywikibot.page.BasePage, info: str) -> None:
        """Initializer."""
        self.info = info
        super().__init__(page)


class SpamblacklistError(PageSaveRelatedError):

    """Page save failed because MediaWiki detected a blacklisted spam URL."""

    message = ('Edit to page {title} rejected by spam filter due to '
               'content:\n{url}')

    def __init__(self, page: pywikibot.page.BasePage, url: str) -> None:
        """Initializer."""
        self.url = url
        super().__init__(page)


class TitleblacklistError(PageSaveRelatedError):

    """Page save failed because MediaWiki detected a blacklisted page title."""

    message = 'Page {} is title-blacklisted.'


class ClientError(Error):

    """Got unexpected server response due to client issue.

    .. versionadded:: 8.1
    """


class Client414Error(ClientError):

    """Server returned with HTTP 414 code.

    .. versionadded:: 8.1
    """


class ServerError(Error):

    """Got unexpected server response."""


class FatalServerError(ServerError):

    """A fatal server error will not be corrected by resending the request."""


class Server504Error(ServerError):

    """Server timed out with HTTP 504 code."""


class CaptchaError(Error):

    """Captcha is asked and config.solve_captcha == False."""


class AutoblockUserError(Error):

    """Requested action on a virtual autoblock user not valid.

    The class AutoblockUserError is an exception that is raised whenever
    an action is requested on a virtual autoblock user that's not
    available for him (i.e. roughly everything except unblock).
    """


class TranslationError(Error, ImportError):

    """Raised when no correct translation could be found.

    Inherits from ImportError, as this exception is now used where
    previously an ImportError would have been raised, and may have been
    caught by scripts as such.
    """


class UserRightsError(Error):

    """Insufficient user rights to perform an action."""


class HiddenKeyError(UserRightsError, KeyError):

    """Insufficient user rights to view the hidden key."""


class NotEmailableError(PageRelatedError):

    """This user is not emailable."""

    message = '{} is not emailable.'


class PageInUseError(Error):

    """Page cannot be reserved for writing due to existing lock."""


class WikiBaseError(Error):

    """Wikibase related error."""


class NoWikibaseEntityError(WikiBaseError):

    """This entity doesn't exist."""

    def __init__(self, entity: pywikibot.page.WikibaseEntity) -> None:
        """Initializer.

        :param entity: Wikibase entity
        """
        super().__init__(
            f"Entity '{entity.id}' doesn't exist on {entity.repo}")
        self.entity = entity


class CoordinateGlobeUnknownError(WikiBaseError, NotImplementedError):

    """This globe is not implemented yet in either WikiBase or pywikibot."""


class EntityTypeUnknownError(WikiBaseError):

    """The requested entity type is not recognised on this site."""


class TimeoutError(Error):

    """Request failed with a timeout error."""


class MaxlagTimeoutError(TimeoutError):

    """Request failed with a maxlag timeout error."""


class ApiNotAvailableError(Error):

    """API is not available, e.g. due to a network error or configuration."""


wrapper = ModuleDeprecationWrapper(__name__)
wrapper.add_deprecated_attr(
    'Server414Error', Client414Error, since='8.1.0')



"""Abstract base user interface module.

.. versionadded:: 6.2
"""
#
# (C) Pywikibot team, 2021
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import sys
from abc import ABC, abstractmethod
from typing import Any


class ABUIC(ABC):

    """Abstract base user interface class.

    Every user interface should derive from it to ensure that all
    required methods are implemented.

    .. versionadded:: 6.2
    """

    def argvu(self) -> list[str]:
        """Return copy of sys.argv.

        Assigned to pywikibot.argvu in bot module
        """
        return list(sys.argv)

    @abstractmethod
    def flush(self) -> None:
        """Flush cached output.

        May be passed to atexit.register() to flush any ui cache.
        """

    @abstractmethod
    def init_handlers(self, *args, **kwargs) -> None:
        """Initialize the handlers for user output.

        Called in bot.init_handlers().
        """

    @abstractmethod
    def input(self, *args, **kwargs) -> str:
        """Ask the user a question and return the answer.

        Called by bot.input().
        """
        if args:
            return input(args[0])
        return input()

    @abstractmethod
    def input_choice(self, *args, **kwargs) -> int | str:
        """Ask the user and returns a value from the options.

        Called by bot.input_choice().
        """
        return self.input()

    @abstractmethod
    def input_list_choice(self, *args, **kwargs) -> Any:
        """Ask the user to select one entry from a list of entries.

        Called by bot.input_list_choice().
        """
        return self.input()

    @abstractmethod
    def output(self, *args, **kwargs) -> None:
        """Output text to a stream."""
        print(*args, **kwargs)  # noqa: T201
